from osbot_utils.utils.Dev import pprint

from osbot_utils.utils.Lists import list_index_by

from osbot_aws.aws.route_53.Route_53 import Route_53
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class Route_53__Hosted_Zone(Kwargs_To_Self):
    route_53       : Route_53
    hosted_zone_id : str

    def a_records(self):
        records = self.record_sets(group_by='Type').get('A')
        return list_index_by(records, 'Name')

    def create_new_a_record(self, record_name , alias_target, hosted_zone_id, alias_hosted_zone_id):
        create_kwargs = dict(record_name            = record_name          ,
                             record_type            = 'A'                  ,
                             alias_target               = alias_target     ,
                             hosted_zone_id         = hosted_zone_id       ,
                             alias_hosted_zone_id   = alias_hosted_zone_id )

        response = self.route_53.record_set__upsert(**create_kwargs)
        return response.get('ChangeInfo')

    def info(self):
        raw_hosted_zone = self.route_53.hosted_zone(hosted_zone_id=self.hosted_zone_id)
        return raw_hosted_zone.get('HostedZone')

    def record_set(self, record_name, record_type='A'):
        return self.route_53.record_set(hosted_zone_id=self.hosted_zone_id, record_name=record_name, record_type=record_type)

    def delete_record_set__a__to_alias(self, dns_entry, alias_dns_name, alias_hosted_zone_id):
        kwargs = dict(hosted_zone_id       = self.hosted_zone_id ,
                      record_name          = dns_entry         ,
                      record_type          = 'A'                 ,
                      alias_dns_name       = alias_dns_name      ,
                      alias_hosted_zone_id = alias_hosted_zone_id)
        return self.route_53.record_set__delete__alias(**kwargs).get('ChangeInfo')

    def record_sets(self, **kwargs):
        return self.route_53.record_sets(hosted_zone_id=self.hosted_zone_id, **kwargs)

    def check_dns_answer(self, record_name, record__type='A'):
        return self.route_53.check_dns_answer(hosted_zone_id=self.hosted_zone_id, record_name=record_name, record_type=record__type)







import re
from functools                               import wraps
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Dev                   import pprint


def capture_iam_exception(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return_value = None
        with Capture_IAM_Exception() as _:
            return_value = func(*args, **kwargs)
        if _.permission_required.get('status')=='ok':
            print()
            pprint('****** IAM PERMISSION REQUIRED ******')
            del _.permission_required['status']                         # this value can be consuming
            pprint(_.permission_required)
        return return_value

    return wrapper

class Capture_IAM_Exception(Kwargs_To_Self):
    permission_required : dict

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.permission_required = { 'event'     : 'no IAM exception triggered' ,
                                     'status'    : 'error'                  }
        # pprint(exc_type)
        # if exc_type is ClientError:
        if exc_val:
            error_code = exc_val.response.get('Error', {}).get('Code')
            if error_code == 'AccessDeniedException':
                error_message = exc_val.response['Error']['Message']

                user_arn_match = re.search(r"User: arn:aws:iam::(\d+):user/([^ ]+) is not authorized", error_message)
                required_permission_match = re.search(r"because no identity-based policy allows the ([^:]+):([^ ]+) action", error_message)

                if user_arn_match and required_permission_match:
                    account_id, user = user_arn_match.groups()
                    service, action = required_permission_match.groups()

                    self.permission_required = { 'event'     : 'IAM exception triggered' ,
                                                 'status'    : 'ok'                      ,
                                                 'service'   : service                   ,
                                                 'action'    : action                    ,
                                                 'account_id': account_id                ,
                                                 'user'      : user                      }

                return True
        return False



from functools                          import cache
from osbot_aws.aws.s3.S3                import S3
from osbot_aws.aws.iam.IAM_Assume_Role  import IAM_Assume_Role
from osbot_utils.utils.Env import load_dotenv


class S3__with_temp_role(S3):
    temp_service_name   : str = "s3"
    temp_role__services : str = ["s3"]
    temp_role__action   : str = "*"
    temp_role__resource : str = "*"
    temp_role__name     : str = 'osbot__temp_role_for__test_S3'
    temp_role__recreate : bool = False

    def iam_assume_role(self):
        load_dotenv()
        action        = self.temp_role__action
        resource      = self.temp_role__resource
        role_name     = self.temp_role__name

        policies_to_add = []
        for temp_role__service in self.temp_role__services:
            policies_to_add.append(dict(service=temp_role__service, action=action, resource=resource))

        iam_assume_role = IAM_Assume_Role(role_name=role_name, policies_to_add=policies_to_add)
        return iam_assume_role

    def create_role(self, create_credentials=True):
        role_recreate   = self.temp_role__recreate
        iam_assume_role = self.iam_assume_role()
        iam_assume_role.create_role(recreate=role_recreate, create_credentials=create_credentials)
        return iam_assume_role

    
    @cache
    def client(self):
        iam_assume_role = self.create_role()
        service        = self.temp_service_name
        return iam_assume_role.boto3_client(service_name=service)

    def sts(self):
         return self.iam_assume_role().aws__client__sts()            # this is an STS object to the current identity (not the one used to create it)

    def temp_role__iam_reset_credentials(self):
        iam_assume_role = IAM_Assume_Role(role_name=self.temp_role__name)
        return iam_assume_role.credentials_reset()


import sys
from botocore.exceptions                                    import ClientError, NoCredentialsError
from osbot_utils.decorators.methods.remove_return_value     import remove_return_value
from osbot_utils.utils.Misc                                 import random_string
from osbot_utils.utils.Status                               import status_ok, status_error, status_warning
from osbot_utils.decorators.methods.cache_on_self           import cache_on_self
from osbot_aws.AWS_Config                                   import AWS_Config

class STS:
    """
    Helper methods for the AWS STS (Security Token Service)
    """
    def __init__(self, print_error_message=True):
        self.print_error_message = print_error_message                  # todo: need to find a better way to have the bad credentials errors showing up in IDE when development

    @cache_on_self
    def client(self):
        from osbot_aws.apis.Session import Session                      # recursive dependency
        return Session().client('sts')

    @remove_return_value('ResponseMetadata')
    def assume_role(self, role_arn, role_session_name=None, duration_seconds=900):
        role_session_name = role_session_name or random_string(prefix='osbot_sts_session_')
        kwargs = { "DurationSeconds" : duration_seconds , # minumum 900 seconds (15 minutes) manximum 43200 seconds (15 hour)
                    "RoleArn"        : role_arn         ,
                    "RoleSessionName": role_session_name or ''}
        return self.client().assume_role(**kwargs)

    def current_account_id(self):
        return self.caller_identity_account()

    def current_region_name(self):
        return self.client().meta.region_name

    @remove_return_value('ResponseMetadata')
    def caller_identity(self):
        try:
            return self.client().get_caller_identity()
        except ClientError as client_error:
            print(f"[error][STS] - ClientError {client_error}")                 # todo: add better logging support
        except NoCredentialsError as no_credentials_error:
            print(f"[error][STS] - NoCredentialsError {no_credentials_error}")  # todo: add better logging support
        except Exception as exception:
            print(f"[error][STS] - Exception {exception}")                      # todo: add better logging support
        return {}

    def caller_identity_account(self):
        return self.caller_identity().get('Account')

    def caller_identity_arn(self):
        return self.caller_identity().get('Arn')

    def caller_identity_user(self):
        arn = self.caller_identity().get('Arn')
        if ':user/' in arn:
            return arn.split(':user/')[-1]
        elif ':assumed-role/' in arn:
            return arn.split(':assumed-role/')[-1].split('/')[1]
        else:
            raise ValueError("ARN does not contain a user or assumed-role")

    def check_aws_session(self):
        try:
            caller_arn = self.caller_identity().get('Arn')
            return status_ok()
        except ClientError as client_error:
            exception = client_error.response.get('Error')
            error_code    = exception.get("Code")
            error_message = exception.get('Message')
            message = (f"\tCurrent Security credentials are invalid! \n\n"
                       f"\tPlease check: current environment values, "
                       f"profile name or value in your ~/.aws/credentials file"
                       f"\n\n"
                       f"\tBoto3 Error message: {error_code}  {error_message}")
            return status_error(message=message, error=client_error)

    @cache_on_self
    def check_current_session_credentials(self, raise_exception=True):          # todo: see if there is a faster way to do this, at the moment it takes about 500ms which is quite a lot
        if AWS_Config().dev_skip_aws_key_check() == "True":
            return status_warning(message="check was skipped")
        result = self.check_aws_session()
        if result.get('status') == 'error':
            self.print_bad_credentials_exception(result.get('message'))
            #if exit_on_error:                                                  # todo: see better way to do this
            #    self.end_process_bad_credentials_exception()
            if raise_exception:
                self.raise_bad_credentials_exception(result.get('message'))
        return result

    def print_bad_credentials_exception(self, message):
        if self.print_error_message:
            print("\n*********** OSBot_AWS ***********\n")             # todo move to OSBot_AWS Logging class
            print(f"{message}")
            print("\n*********** OSBot_AWS ***********\n")

    def raise_bad_credentials_exception(self, message):
        sys.trackbacklimit = 0
        raise Exception(message)

    # def end_process_bad_credentials_exception(self):          # todo: figure out how to do this (since this always throws an exception stack trace, at least in testing)
    #     quit(0)
    #     exit(0)

def check_current_aws_credentials(raise_exception=True):
    return STS().check_current_session_credentials(raise_exception=raise_exception)

def current_aws_region():
    return STS().current_region_name()


from osbot_aws.helpers.Lambda_Layer_Create import Lambda_Layer_Create


class Lambda_Layers_OSBot:

    PACKAGE__OSBOT_UTILS = "git+https://github.com/owasp-sbot/OSBot-Utils.git@dev"
    PACKAGE__OSBOT_AWS   = 'git+https://github.com/owasp-sbot/OSBot-AWS.git@dev'

    def __init(self):
        pass

    def create__fastapi(self):
        packages = ['fastapi', 'uvicorn']
        with Lambda_Layer_Create('layer_for__fastapi') as _:
            _.add_packages(packages)
            _.recreate()
            return _.arn_latest()

    def create__flask(self):
        packages = ['flask', 'serverless_wsgi']
        with Lambda_Layer_Create('layer_for__flask') as _:
            _.add_packages(packages)
            _.create()
            return _.arn_latest()

    def create__llms(self):
        packages = ['openai']
        with Lambda_Layer_Create('layer_for__llms') as _:
            _.delete_local_layer_folder()
            _.add_packages(packages)
            _.recreate()
            return _.arn_latest()

    def create__mangum(self):       # todo: refactor all these create__xxx methods
        packages = ['mangum']
        with Lambda_Layer_Create('layer_for__mangum') as _:
            _.add_packages(packages)
            _.recreate()
            return _.arn_latest()

    def create__osbot_aws(self):
        packages = [self.PACKAGE__OSBOT_UTILS   ,
                    self.PACKAGE__OSBOT_AWS     ,
                    'python-dotenv'             ]
        with Lambda_Layer_Create('layer_for__osbot_aws') as _:
            _.installed_packages_reset()                                # todo: add way to only reset a couple specific packages
            _.add_packages(packages)
            _.recreate()
            return _.arn_latest()

    def create__osbot_utils(self):
        packages = [self.PACKAGE__OSBOT_UTILS, 'python-dotenv' ]
        with Lambda_Layer_Create('layer_for__osbot_utils') as _:
            _.add_packages(packages)
            _.recreate()
            return _.arn_latest()

    def fastapi(self):
        return Lambda_Layer_Create('layer_for__fastapi').arn_latest()

    def flask(self):
        return Lambda_Layer_Create('layer_for__flask').arn_latest()

    def llms(self):
        return Lambda_Layer_Create('layer_for__llms').arn_latest()

    def mangum(self):
        return Lambda_Layer_Create('layer_for__mangum').arn_latest()

    def osbot_aws(self):
        return Lambda_Layer_Create('layer_for__osbot_aws').arn_latest()

    def osbot_utils(self):
        with Lambda_Layer_Create('layer_for__osbot_utils') as _:
            _.add_package(self.PACKAGE__OSBOT_UTILS)
            _.add_package('python-dotenv')
            return _.create()                                   # will return _.arn_latest() if already exists
        #return layer_osbot_utils.installed_packages_names()




from functools import wraps

from osbot_aws.aws.sts.STS import STS


class aws_inject:

    """injects a number of AWS Specific values"""
    def __init__(self, fields):
        self.fields = fields                                        # field to inject

    def __call__(self, function):
        @wraps(function)                                            # makes __name__ work ok
        def wrapper(*args,**kwargs):                                # wrapper function
            for field in self.fields.split(','):                    # split value provided by comma
                if field == 'region'    : kwargs[field] = STS().current_region_name()
                if field == 'account_id': kwargs[field] = STS().current_account_id()
            return function(*args,**kwargs)
        return wrapper


from osbot_aws.aws.apigateway.apigw_dydb.WS__DyDB           import WS__DyDB
from osbot_aws.aws.apigateway.apigw_dydb.WS__Handle_Lambda  import WS__Handle_Lambda


class WS__Handle_Events(WS__Handle_Lambda):
    dydb_ws    : WS__DyDB
    source     : str        = 'WS__Handle_Events'


    def route__connect(self, event, connection_id):
        kwargs   = dict(connection_id = connection_id ,
                        source        = self.source   ,
                        lambda_event  = event         )
        self.dydb_ws.register_connection(**kwargs)
        return {'statusCode': 200, 'body': 'Connected.'}

    def route_default(self, event, connection_id):
        kwargs   = dict(connection_id = connection_id ,
                        source        = self.source   ,
                        lambda_event  = event         )
        return self.dydb_ws.handle_route(**kwargs)

    def route_disconnect(self, event, connection_id):
        kwargs   = dict(connection_id = connection_id ,
                        source        = self.source   ,
                        lambda_event  = event         )
        self.dydb_ws.register_disconnect(**kwargs)
        return {'statusCode': 200, 'body': 'Disconnected.'}




from osbot_aws.AWS_Config               import AWS_Config
from osbot_aws.aws.s3.S3                import S3
from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Files import is_file

FORMAT__TEMP_BUCKET_NAME = S3_TEMP_BUCK = '{account_id}--osbot--temp--{region_name}'

class S3__On_Temp_Bucket(Type_Safe):
    aws_config: AWS_Config
    s3        : S3

    def add_file(self, path_to_file, s3_key=None):
        return self.s3.file_upload_to_key(file=path_to_file, bucket=self.bucket_name(), key=s3_key)

    def create_file__from_string(self, s3_key, file_contents):
            return self.s3.file_create_from_string(bucket=self.bucket_name(), key=s3_key, file_contents=file_contents)

    def create_file__from_bytes(self, s3_key, file_bytes):
        return self.s3.file_create_from_bytes(bucket=self.bucket_name(), key=s3_key, file_bytes=file_bytes)

    def bucket_exists(self):
        return self.s3.bucket_exists(self.bucket_name())

    def bucket_name(self):
        account_id = self.aws_config.account_id()
        region_name = self.aws_config.region_name()
        return FORMAT__TEMP_BUCKET_NAME.format(account_id=account_id, region_name=region_name)

    def file_contents(self, s3_key):
        return self.s3.file_contents(self.bucket_name(), s3_key)

    def file_delete(self, s3_key):
        return self.s3.file_delete(self.bucket_name(), s3_key)

    def file_exists(self, s3_key):
        return self.s3.file_exists(self.bucket_name(), s3_key)



from osbot_aws.aws.dynamo_db.domains.DyDB__Table     import DyDB__Table
from osbot_aws.aws.dynamo_db.models.DyDB__Document   import DyDB__Document
from osbot_aws.aws.dynamo_db.models.Resource__Config import Resource__Config
from osbot_utils.decorators.methods.cache_on_self    import cache_on_self

KEY_NAME__TABLE__RESOURCES = 'resource'

class DyDB__Table__Resources(DyDB__Table):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.key_name =  KEY_NAME__TABLE__RESOURCES

    def resources_types(self):
        resources_types = {'config': Resource__Config}
        return resources_types

    @cache_on_self
    def resources(self):
        resources = {}
        for resource_name, resource_type in self.resources_types().items():
            document = {KEY_NAME__TABLE__RESOURCES: resource_name}
            resource = resource_type(table =self, document=document)
            resources[resource_name] = resource
        return resources

    def resource(self, resource_name) -> DyDB__Document:
        return self.resources().get(resource_name)


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class Anthropic__Claude_Messages_API(Kwargs_To_Self):
    anthropic_version   : str = "bedrock-2023-05-31"
    max_tokens          : int   = 1024
    messages            : list
    model_id            : str
    #system              : str
    temperature         : float = 0.0
    top_p               : int   = 1
    top_k               : int   = 0
    stop_sequences      : list

    def body(self, messages, system=''):
        return dict(anthropic_version = self.anthropic_version  ,
                    max_tokens        = self.max_tokens         ,
                    messages          = messages                ,
                    temperature       = self.temperature        ,
                    top_p             = self.top_p              ,
                    top_k             = self.top_k              ,
                    stop_sequences    = self.stop_sequences     ,
                    system           = system

                    )


import  boto3
from botocore.config            import Config
from botocore.exceptions        import ClientError
from    botocore.session        import get_session

from osbot_utils.base_classes.Kwargs_To_Self            import Kwargs_To_Self
from osbot_utils.context_managers.capture_duration import print_duration
from osbot_utils.decorators.methods.cache               import cache
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_utils.utils.Status                           import status_ok, status_error
from osbot_aws.AWS_Config                               import aws_config
from osbot_aws.exceptions.Session_Bad_Credentials       import Session_Bad_Credentials
from osbot_aws.exceptions.Session_Client_Creation_Fail  import Session_Client_Creation_Fail
from osbot_aws.exceptions.Session_No_Credentials        import Session_No_Credentials


class Session(Kwargs_To_Self):                  # todo: refactor to AWS_Session so it doesn't clash with boto3.Session object
    account_id      = None
    config          = None
    profile_name    = None
    region_name     = None                      # keep a copy of this value


    def boto_session(self) -> boto3.Session:
        return get_session()

    @cache
    def botocore_session(self, aws_access_key_id=None , aws_secret_access_key=None , aws_session_token= None, region_name= None, botocore_session=None, profile_name=None ):
        kwargs = {  "aws_access_key_id"     : aws_access_key_id     ,
                    "aws_secret_access_key" : aws_secret_access_key ,
                    "aws_session_token"     : aws_session_token     ,
                    "region_name"           : region_name           ,
                    "botocore_session"      : botocore_session      ,
                    "profile_name"          : profile_name          }
        session = boto3.Session(**kwargs)
        return session

    def credentials(self, session=None):
        session = session or self.botocore_session()
        credentials  = session.get_credentials()
        if credentials is None:
            raise Session_No_Credentials()
        return { "access_key"   : credentials.access_key ,
                 "method"       : credentials.method     ,
                 "profile"      : session.profile_name   ,
                 "secret_key"   : credentials.secret_key ,
                 "token"        : credentials.token      }

    def credentials_ok(self):                                   # see if we still need this method
        return self.credentials() is not None
        #sts = self.botocore_session().client('sts')
        #return sts.get_caller_identity()
        # try:
        # except ClientError as exception:
        #
        #     print(exception.response['Error']['Code'] == 'ResourceAlreadyExistsException')
        #     print(exception.response['Error']['Code'])
        #     raise

    def caller_identity(self, session=None):
        session = session or self.botocore_session()
        try:
            return session.client('sts').get_caller_identity()             # few cases where we want to call this directly (since the STS class will use the session() method from this class
        except ClientError as client_error:
            exception = client_error.response.get('Error')
            raise Session_Bad_Credentials(exception)

    @cache_on_self
    def session(self) -> boto3.Session:
        return boto3.Session()

    def session_default(self):
        return get_session()

    def client(self, service_name, **kwargs):
        status = self.client_boto3(service_name, **kwargs)
        if status.get('status') == 'ok':
            client = status.get('data',{}).get('client')
            return client
        else:
            raise Session_Client_Creation_Fail(status=status)

    def client_boto3(self,service_name, aws_access_key_id=None, aws_secret_access_key=None, region_name=None, config=None, endpoint_url=None, profile_name=None):                   # todo: refactor add this to resource_boto3
        try:
            self.config       = config       or self.config       or self.default_config                ()
            self.region_name  = region_name  or self.region_name  or aws_config.aws_session_region_name ()
            self.profile_name = profile_name or self.profile_name or aws_config.aws_session_profile_name()

            client_kwargs = dict( service_name = service_name , config = self.config)

            if aws_access_key_id    : client_kwargs['aws_access_key_id'    ] = aws_access_key_id
            if aws_secret_access_key: client_kwargs['aws_secret_access_key'] = aws_secret_access_key
            if endpoint_url         : client_kwargs['endpoint_url'         ] = endpoint_url
            if self.region_name     : client_kwargs['region_name'          ] = self.region_name

            if self.profile_name and self.profile_name in self.profiles():                                                  # seeing if this is a more efficient way to get the data
                session = boto3.Session(profile_name=self.profile_name, region_name=self.region_name)      # tried to pass this params but had side effects: , botocore_session=self.boto_session()
                client  = session.client(**client_kwargs)
                message = f"Created client_boto3 session using profile: {self.profile_name}"
            else:
                client = boto3.client(**client_kwargs)
                session = None
                message = "Created client_boto3 session using environment variables"             # todo: see if this is 100% correct, or if there are other ways credentials can be found inside boto3 client

            data = {"client" : client, "session" : session }
            return status_ok(message=message, data=data)
        except Exception as error:

            message = f"client_boto3 failed with error: {error}"
            print(message)
            return status_error(message=message, data=None, error=error)

    @cache_on_self
    def default_config(self):
        retries = {
            'max_attempts': 1,  # set retry to 1  (when used 0 it was causing some side effects)
            'mode': 'standard'  # Standard retry mode
        }
        kwargs = dict(retries=retries, signature_version='s3v4')
        return Config(**kwargs)

    def profiles(self):
        return self.session().available_profiles

    def profiles_map(self):
        return get_session()._build_profile_map()

        #return self.boto_session()._build_profile_map()

    def resource_boto3(self,service_name, profile_name=None, region_name=None):                 # todo: refactor with client_boto3
        try:
            profile_name = profile_name or aws_config.aws_session_profile_name()
            region_name  = region_name  or aws_config.aws_session_region_name()

            profiles = get_session()._build_profile_map()
            if profile_name in profiles:
                session = boto3.Session(profile_name=profile_name, region_name=region_name)
                return {'status': 'ok', 'resource': session.resource(service_name=service_name) }
            return { 'status' : 'ok', 'resource': boto3.resource(service_name=service_name)}
        except Exception as error:
            return {'status': 'error', 'data': '{0}'.format(error) }

    def resource(self, service_name,profile_name=None, region_name=None):
        return self.resource_boto3(service_name,profile_name,region_name).get('resource')

    # client helpers

    # putting all these here in global location so that we limit the calls to authenticate
    # important when making lots of calls of multiple apis (like when running multiple tests)
    # todo find good way to add support for refreshing these references (for exemple to use a different region)
    # todo: refactor this out of this session object
    @cache
    def client_secrets_manager(self):
        return self.client('secretsmanager')

    # todo: refactor into this class (from IAM.py)
    # def access_key__is_key_working(self, access_key):       # refactor to Sessions class
    #     aws_access_key_id     = access_key.get('AccessKeyId')
    #     aws_secret_access_key = access_key.get('SecretAccessKey')
    #     try:
    #         session = boto3.Session(aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
    #         session.client('sts').get_caller_identity()
    #         return True
    #     except Exception as error:
    #         assert error.__str__() == 'An error occurred (InvalidClientTokenId) when calling the GetCallerIdentity operation: The security token included in the request is invalid.'
    #         return False




import os
import importlib
import site

from osbot_utils.testing.Temp_Folder            import Temp_Folder
from osbot_utils.testing.Temp_Zip import Temp_Zip
from osbot_utils.utils.Process                  import Process
from osbot_aws.apis.Lambda_Layer                import Lambda_Layer
from osbot_aws.AWS_Config                       import AWS_Config
from osbot_aws.apis.Lambda                      import Lambda
from osbot_aws.apis.test_helpers.Temp_Aws_Roles import Temp_Aws_Roles
from osbot_utils.utils.Files                    import folder_copy, Files, folder_not_exists, path_combine, folder_name



class Lambda_Package:
    def __init__(self,lambda_name):
        self.lambda_name   = lambda_name
        self.aws_lambda    = Lambda(self.lambda_name)
        self.s3_bucket     = AWS_Config().lambda_s3_bucket()
        self.s3_key        = f'{AWS_Config().lambda_s3_folder_lambdas()}/{self.lambda_name}.zip'
        self.role_arn      = Temp_Aws_Roles().for_lambda_invocation__role_arn()
        self.tmp_folder    = Files.temp_folder('tmp_lambda_')
        self.layers        = []
        self.env_variables = {}
        (self.aws_lambda.set_s3_bucket   (self.s3_bucket    )
                        .set_s3_key      (self.s3_key       )
                        .set_role        (self.role_arn     )
                        .set_folder_code (self.tmp_folder  ))

    # helper methods
    @staticmethod
    def get_root_folder():
        return Files.path_combine(__file__, '../..')

    # Lambda class wrappers

    def create(self             ): return self.aws_lambda.create()
    def delete(self             ): return self.aws_lambda.delete()
    def invoke(self, params=None): return self.aws_lambda.invoke(params)
    def update(self             ): return self.aws_lambda.update()
    def reset (self             ): return self.aws_lambda.update_lambda_code()              # this will trigger a reset and force cold start on next execution

    # main methods
    def add_layer(self, layer_arn):
        self.aws_lambda.add_layer(layer_arn)

    def add_layers(self, layers_arn):
        for layer_arn in layers_arn:
            self.aws_lambda.add_layer(layer_arn)
        return self

    def add_file(self, source):
        Files.copy(source, self.tmp_folder)

    def add_folder(self, source, ignore=None):
        destination = Files.path_combine(self.tmp_folder, folder_name(source))
        if folder_not_exists(destination):
            folder_copy(source=source, destination=destination,ignore_pattern=ignore)
        self.remove_files('__pycache__')
        return self

    def add_module(self,module_name):
        module_path = importlib.import_module(module_name).__path__[0]     # get folder of module
        self.add_folder(module_path)                                       # add module's folder
        return self

    def add_modules(self, modules_names):
        for module_name in modules_names:
            self.add_module(module_name)
        return self

    def add_root_folder(self):
        self.add_folder(self.get_root_folder())
        return self

    def add_osbot_aws(self):
        self.add_osbot_utils()
        self.add_module('osbot_aws')

    def add_osbot_browser(self):
        self.add_module('osbot_browser')

    def add_osbot_elastic(self):
        self.add_module('osbot_elastic')

    def add_osbot_utils(self):
        self.add_module('dotenv')
        self.add_module('osbot_utils')
        return self

    def arn(self):
        return self.aws_lambda.function_arn()

    def create_layer(self, target_folder, ignore_pattern=None):
        if ignore_pattern is None:
            ignore_pattern = ['*.pyc', '.DS_Store', '__pycache__', '.env']

        layer_name   = folder_name(target_folder)
        lambda_layer = Lambda_Layer(layer_name)
        with Temp_Folder() as _:
            source      = target_folder
            destination = path_combine(_.path() + '/python', folder_name(source))
            folder_copy(source=source, destination=destination, ignore_pattern=ignore_pattern)
            with Temp_Zip(_.path()) as zip_file:
                result = lambda_layer.create_from_zip_file_via_s3(zip_file.path())
                return  result.get('LayerVersionArn')

    def create_layer_from_packages(self, layer_name, packages, target_aws_lambda=True):
        # if ignore_pattern is None:
        #     ignore_pattern = ['*.pyc', '.DS_Store', '__pycache__', '.env']

        with Temp_Folder() as _:
            destination = path_combine(_.path(), 'python')
            for package in packages:
                args = ['install']
                if target_aws_lambda:
                    args.extend(['--platform', 'manylinux1_x86_64', '--only-binary=:all:'])
                args.extend(['-t', destination, package])
                Process.run('pip3', args)
            with Temp_Zip(_.path()) as zip_file:
                lambda_layer = Lambda_Layer(layer_name)
                result       = lambda_layer.create_from_zip_file_via_s3(zip_file.path())
                return result.get('LayerVersionArn')


    def create_layer_from_site_packages(self, target_name, ignore_pattern=None):
        if ignore_pattern is None:
            ignore_pattern = ['*.pyc', '.DS_Store', '__pycache__', '.env']

        with Temp_Folder() as _:
            layer_name   = f'site_packages__{target_name}'
            lambda_layer = Lambda_Layer(layer_name)
            source       = site.getsitepackages()[0]
            destination  = path_combine(_.path(), 'python')
            folder_copy(source=source, destination=destination, ignore_pattern=ignore_pattern)
            with Temp_Zip(_.path()) as zip_file:
                result = lambda_layer.create_from_zip_file_via_s3(zip_file.path())
                return  result.get('LayerVersionArn')

    def get_files(self):
        all_files = []
        for root, dirs, files in os.walk(self.tmp_folder):
            for file in files:
                file_path = Files.path_combine(root,file).replace(self.tmp_folder,'')
                all_files.append(file_path)
        return all_files


    def remove_files(self,pattern):
        for file in self.get_files():
            if pattern in file:
                file_to_delete = Files.path_combine(self.tmp_folder,file[1:])
                Files.delete(file_to_delete)

    def set_handler(self, handler):
        self.aws_lambda.set_handler(handler)
        return self

    def set_env_variable(self, key, value):
        self.aws_lambda.set_env_variable(key, value)
        return self

    def set_env_variables(self, env_variables):
        self.aws_lambda.set_env_variables(env_variables)
        return self

    def set_layers(self, layers):
        self.aws_lambda.set_layers(layers)
        return self

    def set_image_uri(self, image_uri):
        self.aws_lambda.image_uri = image_uri

    def use_lambda_file(self,lambda_file):
        file_path = Files.path_combine(self.get_root_folder(), lambda_file)
        if Files.exists(file_path) is False:
            return { 'status': 'error', 'data': 'could not find lambda file `{0}` in root folder `{1}`'.format(lambda_file, self.get_root_folder())}
        target_file = Files.path_combine(self.tmp_folder, '{0}.py'.format(self.lambda_name))
        Files.copy(file_path,target_file)
        return  { 'status': 'ok', 'file_path': file_path, 'target_file': target_file }

    def uses_image_uri(self):
        return self.aws_lambda.image_uri is not None

    def update_with_root_folder(self,delete_before=False):
        if delete_before:
            self.delete()
        self.add_root_folder()
        self.add_osbot_utils()
        self.update()
        return self

    def update_code(self):
        base_folder = self.lambda_name.split('.').pop(0)                    # will point to the top level module
        self.add_module(base_folder)                                        # add module's folder
        self.add_root_folder()                                              # add osbot-aws folder
        self.add_osbot_utils()                                              # add osbot-utils folder
        self.update()                                                       # zip files up and update lambda
        return self




import boto3

from osbot_aws.aws.dynamo_db.Dynamo_DB import Dynamo_DB
from osbot_utils.testing.Temp_Env_Vars  import Temp_Env_Vars
from unittest                           import TestCase
from osbot_utils.testing.Hook_Method    import Hook_Method
from osbot_utils.utils.Env              import get_env

TEST__AWS_ACCOUNT_ID        = '111122223333'
TEST__AWS_DEFAULT_REGION    = 'eu-west-1'
TEST__AWS_ACCESS_KEY_ID     = 'aaaaaaaaa'
TEST__AWS_SECRET_ACCESS_KEY = 'bbbbbbbbb'
URL_DOCKER__DYNAMODB__LOCAL = 'http://localhost:8000'

class TestCase__Dynamo_DB__Local(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.aws_account_id          = get_env('AWS_ACCOUNT_ID'       , TEST__AWS_ACCOUNT_ID       )         # use existing values if they already exist
        cls.aws_default_region      = get_env('AWS_DEFAULT_REGION'   , TEST__AWS_DEFAULT_REGION   )
        cls.aws_access_key_id       = get_env('AWS_ACCESS_KEY_ID'    , TEST__AWS_ACCESS_KEY_ID    )         # this is neeed by create table
        cls.aws_secret_access_key   = get_env('AWS_SECRET_ACCESS_KEY', TEST__AWS_SECRET_ACCESS_KEY)
        tmp_vars_values             = dict(AWS_ACCOUNT_ID        = cls.aws_account_id       ,           # todo: find better way to do this, since this will impact all code executed in the current test
                                           AWS_DEFAULT_REGION    = cls.aws_default_region   ,
                                           AWS_ACCESS_KEY_ID     = cls.aws_account_id       ,
                                           AWS_SECRET_ACCESS_KEY = cls.aws_secret_access_key)
        cls.tmp_env_vars            = Temp_Env_Vars(env_vars=tmp_vars_values).set_vars()
        cls.hook_method             = Hook_Method(Dynamo_DB, 'client'     )              # todo: see if we also need to hook the main S3 class (from osbot_aws)
        cls.hook_method.mock_call   = cls.dynamo_db__client
        cls.hook_method.wrap()
        cls.dynamo_db               = Dynamo_DB()

    @classmethod
    def tearDownClass(cls):
        cls.hook_method.unwrap()
        cls.tmp_env_vars.restore_vars()

    def dynamo_db__client(self):
        return boto3.client('dynamodb', endpoint_url=URL_DOCKER__DYNAMODB__LOCAL)


from osbot_aws.AWS_Config import AWS_Config
from osbot_aws.apis.Session import Session
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_utils.utils.Files import stream_to_bytes
from osbot_utils.utils.Zip import gz_tar_bytes__file_list, gz_zip_bytes__file_list, zip_bytes__file_list, \
    zip_bytes__extract_to_folder


class Code_Artifact(Kwargs_To_Self):
    aws_config  : AWS_Config
    region_name : str
    @cache_on_self
    def client(self):
        return Session().client('codeartifact', region_name=self.region_name)

    def asset_name__gz(self, package_name, version):
        package_name = package_name.replace('-', '_')           # can't have - in asset_name
        asset_name   = f'{package_name}-{version}.tar.gz'
        return asset_name

    def asset_name__whl(self, package_name, version):
        package_name = package_name.replace('-', '_')
        asset_name   = f'{package_name}-{version}-py3-none-any.whl'
        return asset_name

    def authorization_token(self, domain):
        return self.client().get_authorization_token(domain=domain).get('authorizationToken')

    def domain(self, domain):
        return self.client().describe_domain(domain=domain).get('domain')

    def domains(self):
        return self.client().list_domains().get('domains')

    def package(self, **kwargs):
        return self.client().describe_package(**kwargs).get('package')

    def package_group(self, **kwargs):
        return self.client().describe_package_group(**kwargs).get('packageGroup')

    def package_groups(self, **kwargs):
        return self.client().list_package_groups(**kwargs).get('packageGroups')

    def package_version(self, **kwargs):
        return self.client().describe_package_version(**kwargs).get('packageVersion')

    @remove_return_value(field_name='ResponseMetadata')
    def package_version_asset(self, **kwargs):
        return self.client().get_package_version_asset(**kwargs)

    def package_version_asset__bytes(self, **kwargs):
        response    = self.package_version_asset(**kwargs)
        asset       = response.get('asset')
        asset_bytes = stream_to_bytes(asset)
        return asset_bytes

    def package_version_asset__files__gz(self, domain, repository,package_name, package_version):
        kwargs =dict(domain         = domain           ,
                     repository     = repository       , #'pypi-store'     ,
                     format         = 'pypi'           ,
                     package        = package_name     ,
                     packageVersion = package_version  )
        asset_name               = self.asset_name__gz(package_name, package_version)
        kwargs['packageVersion'] = package_version
        kwargs['asset'         ] = asset_name
        asset_bytes              = self.package_version_asset__bytes(**kwargs)
        asset_files              = gz_tar_bytes__file_list(asset_bytes)
        return asset_files

    def package_version_asset__files__whl(self, domain, repository,package_name, package_version):
        kwargs =dict(domain         = domain           ,
                     repository     = repository       , #'pypi-store'     ,
                     format         = 'pypi'           ,
                     package        = package_name     ,
                     packageVersion = package_version  )
        asset_name               = self.asset_name__whl(package_name, package_version)
        kwargs['packageVersion'] = package_version
        kwargs['asset'         ] = asset_name
        asset_bytes              = self.package_version_asset__bytes(**kwargs)
        asset_files              = zip_bytes__file_list(asset_bytes)
        return asset_files

    def package_version_asset__whl__extract_to_folder(self, target_folder, **kwargs):
        response    = self.package_version_asset(**kwargs)
        asset       = response.get('asset')
        asset_bytes = stream_to_bytes(asset)
        return zip_bytes__extract_to_folder(asset_bytes, target_folder)

    def package_version_assets(self, **kwargs):
        return self.client().list_package_version_assets(**kwargs).get('assets')

    def package_version_create(self, **kwargs):
        return self.client().publish_package_version(**kwargs)

    def package_version_dependencies(self, **kwargs):
        return self.client().list_package_version_dependencies(**kwargs).get('dependencies')

    @remove_return_value('ResponseMetadata')
    def package_versions(self, **kwargs):
        return self.client().list_package_versions(**kwargs)


    def packages(self, domain, repository):
        return self.client().list_packages(domain=domain, repository=repository).get('packages')

    def repository(self, **kwargs):
        return self.client().describe_repository(**kwargs).get('repository')

    def repository_endpoint(self, **kwargs):
        return self.client().get_repository_endpoint(**kwargs).get('repositoryEndpoint')

    def repositories(self):
        return self.client().list_repositories().get('repositories')


from osbot_utils.utils.Misc import random_string

from osbot_aws.helpers.SQS_Queue import SQS_Queue
from osbot_utils.utils import Misc


class Temp_SQS_Queue:
    def __init__(self , queue_name=None, attributes=None, fifo=True):
        self.attributes = attributes
        self.fifo       = fifo
        self.queue      = None
        self.queue_name = queue_name or random_string(prefix='unit_tests_temp_queue_')
        if fifo and self.queue_name.endswith('.fifo') is False:
            self.queue_name += '.fifo'


    def __enter__(self):
        self.queue = SQS_Queue(self.queue_name)
        if self.fifo:
            self.queue.create_fifo(attributes=self.attributes)
        else:
            self.queue.create(attributes=self.attributes)

        return self.queue

    def __exit__(self, type, value, traceback):
        assert self.queue.delete() is True


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

class AI21_Labs_Jurassic_2(Kwargs_To_Self):
    model_id          : str
    prompt            : str
    temperature       : float    = 0.0
    top_p             : float    = 1.0
    max_tokens        : int      = 1024
    stop_sequences    : list
    #countPenalty     : dict #{ "scale": float }
    #presencePenalty  : dict #{"scale": float },
    #frequencyPenalty : dict # {"scale": float }


    def body(self):
        return dict(prompt        = self.prompt         ,
                    temperature   = self.temperature    ,
                    topP          = self.top_p          ,
                    maxTokens     = self.max_tokens     ,
                    stopSequences = self.stop_sequences )



from osbot_utils.testing.Temp_Env_Vars import Temp_Env_Vars


class Temp__Empty__AWS_Credentials(Temp_Env_Vars):

    def __init__(self, **kwargs):
        empty_aws_credentials = dict(AWS_ACCESS_KEY_ID     = '',
                                     AWS_SECRET_ACCESS_KEY = '',
                                     AWS_ACCOUNT_ID        = '',
                                     AWS_DEFAULT_REGION    = '')
        kwargs['env_vars'] = empty_aws_credentials
        super().__init__(**kwargs)


## NOT WORKING (TO TRY Later)

# import sys, os, base64, datetime, hashlib, hmac
# import requests
# from osbot_aws.apis.IAM import IAM
#
# # based on the example from https://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
#
#
# class Lambda_Http:
#     @aws_inject('region')
#     def __init__(self, lambda_name, region):
#         self.lambda_name = lambda_name
#         self.region      = region
#
#     def sign(self, key, msg):
#         return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()
#
#     def getSignatureKey(self, key, dateStamp, regionName, serviceName):
#         kDate    = self.sign(('AWS4' + key).encode('utf-8'), dateStamp)
#         kRegion  = self.sign(kDate, regionName)
#         kService = self.sign(kRegion, serviceName)
#         kSigning = self.sign(kService, 'aws4_request')
#         return kSigning
#
#     @catch
#     def make_request(self):
#
#         # ************* REQUEST VALUES *************
#         # method = 'POST'
#         # service = 'dynamodb'
#         # host = 'dynamodb.us-west-2.amazonaws.com'
#         # region = 'us-west-2'
#         # endpoint = 'https://dynamodb.us-west-2.amazonaws.com/'
#
#         method   = 'POST'
#         service  = 'lambda'
#         host     = f'lambda.{self.region}.amazonaws.com'
#         region   = self.region
#         endpoint = f'https://{host}/'
#
#         # POST requests use a content type header. For DynamoDB,
#         # the content is JSON.
#         content_type = 'application/x-amz-json-1.0'
#         # DynamoDB requires an x-amz-target header that has this format:
#         #     DynamoDB_<API version>.<operationName>
#         amz_target = 'DynamoDB_20120810.CreateTable'
#
#         # Request parameters for CreateTable--passed in a JSON block.
#         request_parameters =  '{'
#         request_parameters +=  '"KeySchema": [{"KeyType": "HASH","AttributeName": "Id"}],'
#         request_parameters +=  '"TableName": "TestTable","AttributeDefinitions": [{"AttributeName": "Id","AttributeType": "S"}],'
#         request_parameters +=  '"ProvisionedThroughput": {"WriteCapacityUnits": 5,"ReadCapacityUnits": 5}'
#         request_parameters +=  '}'
#
#         # Key derivation functions. See:
#         # http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
#         def sign(key, msg):
#             return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()
#
#         def getSignatureKey(key, date_stamp, regionName, serviceName):
#             kDate = sign(('AWS4' + key).encode('utf-8'), date_stamp)
#             kRegion = sign(kDate, regionName)
#             kService = sign(kRegion, serviceName)
#             kSigning = sign(kService, 'aws4_request')
#             return kSigning
#
#         # Read AWS access key from env. variables or configuration file. Best practice is NOT
#         # to embed credentials in code.
#
#         session = IAM().session()
#         credentials = session.get_credentials()
#         access_key = credentials.access_key
#         secret_key = credentials.secret_key
#
#         #access_key = os.environ.get('AWS_ACCESS_KEY_ID')
#         #secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
#         if access_key is None or secret_key is None:
#             print('No access key is available.')
#             sys.exit()
#
#         # Create a date for headers and the credential string
#         t = datetime.datetime.utcnow()
#         amz_date = t.strftime('%Y%m%dT%H%M%SZ')
#         date_stamp = t.strftime('%Y%m%d') # Date w/o time, used in credential scope
#
#
#         # ************* TASK 1: CREATE A CANONICAL REQUEST *************
#         # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html
#
#         # Step 1 is to define the verb (GET, POST, etc.)--already done.
#
#         # Step 2: Create canonical URI--the part of the URI from domain to query
#         # string (use '/' if no path)
#         canonical_uri = '/'
#
#         ## Step 3: Create the canonical query string. In this example, request
#         # parameters are passed in the body of the request and the query string
#         # is blank.
#         canonical_querystring = ''
#
#         # Step 4: Create the canonical headers. Header names must be trimmed
#         # and lowercase, and sorted in code point order from low to high.
#         # Note that there is a trailing \n.
#         canonical_headers = 'content-type:' + content_type + '\n' + 'host:' + host + '\n' + 'x-amz-date:' + amz_date + '\n' + 'x-amz-target:' + amz_target + '\n'
#
#         # Step 5: Create the list of signed headers. This lists the headers
#         # in the canonical_headers list, delimited with ";" and in alpha order.
#         # Note: The request can include any headers; canonical_headers and
#         # signed_headers include those that you want to be included in the
#         # hash of the request. "Host" and "x-amz-date" are always required.
#         # For DynamoDB, content-type and x-amz-target are also required.
#         signed_headers = 'content-type;host;x-amz-date;x-amz-target'
#
#         # Step 6: Create payload hash. In this example, the payload (body of
#         # the request) contains the request parameters.
#         payload_hash = hashlib.sha256(request_parameters.encode('utf-8')).hexdigest()
#
#         # Step 7: Combine elements to create canonical request
#         canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash
#
#
#         # ************* TASK 2: CREATE THE STRING TO SIGN*************
#         # Match the algorithm to the hashing algorithm you use, either SHA-1 or
#         # SHA-256 (recommended)
#         algorithm = 'AWS4-HMAC-SHA256'
#         credential_scope = date_stamp + '/' + region + '/' + service + '/' + 'aws4_request'
#         string_to_sign = algorithm + '\n' +  amz_date + '\n' +  credential_scope + '\n' +  hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
#
#         # ************* TASK 3: CALCULATE THE SIGNATURE *************
#         # Create the signing key using the function defined above.
#         signing_key = getSignatureKey(secret_key, date_stamp, region, service)
#
#         # Sign the string_to_sign using the signing_key
#         signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'), hashlib.sha256).hexdigest()
#
#
#         # ************* TASK 4: ADD SIGNING INFORMATION TO THE REQUEST *************
#         # Put the signature information in a header named Authorization.
#         authorization_header = algorithm + ' ' + 'Credential=' + access_key + '/' + credential_scope + ', ' +  'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature
#
#         # For DynamoDB, the request can include any headers, but MUST include "host", "x-amz-date",
#         # "x-amz-target", "content-type", and "Authorization". Except for the authorization
#         # header, the headers must be included in the canonical_headers and signed_headers values, as
#         # noted earlier. Order here is not significant.
#         # # Python note: The 'host' header is added automatically by the Python 'requests' library.
#         headers = {'Content-Type':content_type,
#                    'X-Amz-Date':amz_date,
#                    'X-Amz-Target':amz_target,
#                    'Authorization':authorization_header}
#
#
#         # ************* SEND THE REQUEST *************
#         print('\nBEGIN REQUEST++++++++++++++++++++++++++++++++++++')
#         print('Request URL = ' + endpoint)
#
#         r = requests.post(endpoint, data=request_parameters, headers=headers)
#
#         print('\nRESPONSE++++++++++++++++++++++++++++++++++++')
#         print('Response code: %d\n' % r.status_code)
#         print(r.text)
#
#     def make_request__2(self):
#         # request_url = f'https://lambda.{self.region}.amazonaws.com/2015-03-31/functions/{self.lambda_name}/invocations'
#         method             = 'GET'
#         service            = 'lambda'
#         host               = f'lambda.{self.region}.amazonaws.com'
#         region             = self.region
#         canonical_uri      = f'/2015-03-31/functions/{self.lambda_name}/invocations'
#         endpoint           = f'https://lambda.{self.region}.amazonaws.com{canonical_uri}'
#         request_parameters = ''
#
#         session = IAM().session()
#         credentials = session.get_credentials()
#         access_key = credentials.access_key
#         secret_key = credentials.secret_key
#
#         t             = datetime.datetime.utcnow()
#         amzdate       = t.strftime('%Y%m%dT%H%M%SZ')
#         datestamp     = t.strftime('%Y%m%d')
#
#         canonical_querystring = request_parameters
#         canonical_headers = 'host:' + host + '\n' + 'x-amz-date:' + amzdate + '\n'
#         signed_headers = 'host;x-amz-date'
#         payload_hash = hashlib.sha256(('').encode('utf-8')).hexdigest()
#         canonical_request = method + '\n' + canonical_uri + '\n' + canonical_querystring + '\n' + canonical_headers + '\n' + signed_headers + '\n' + payload_hash
#         algorithm = 'AWS4-HMAC-SHA256'
#         credential_scope = datestamp + '/' + region + '/' + service + '/' + 'aws4_request'
#         string_to_sign = algorithm + '\n' + amzdate + '\n' + credential_scope + '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()
#         signing_key = self.getSignatureKey(secret_key, datestamp, region, service)
#         signature = hmac.new(signing_key, (string_to_sign).encode('utf-8'), hashlib.sha256).hexdigest()
#         authorization_header = algorithm + ' ' + 'Credential=' + access_key + '/' + credential_scope + ', ' + 'SignedHeaders=' + signed_headers + ', ' + 'Signature=' + signature
#         headers = {'x-amz-date': amzdate, 'Authorization': authorization_header}
#         request_url = endpoint + '?' + canonical_querystring
#
#         print('\nBEGIN REQUEST++++++++++++++++++++++++++++++++++++')
#         print('Request URL = ' + request_url)
#
#
#         #return request_url
#
#         r = requests.get(request_url, headers=headers)
#
#         print('\nRESPONSE++++++++++++++++++++++++++++++++++++')
#         print('Response code: %d\n' % r.status_code)
#         print(r.text)
#



from functools               import wraps
from datetime                import datetime
from osbot_utils.utils.Files import folder_create, path_combine
from osbot_utils.utils.Json import json_save


def save_event_in_tmp_folder(event):
    now_str       = datetime.now().strftime("%Y-%m-%d__%H-%M-%S__%f")
    target_folder = folder_create('/tmp/event_logs')
    file_name     = f'{now_str}.json'
    target_file = path_combine(target_folder, file_name)

    return json_save(target_file, event)

def lambda_save_event(function):
    @wraps(function)
    def wrapper(event, context=None):
        save_event_in_tmp_folder(event)
        return function(event, context)
    return wrapper


from functools import wraps
from os import getenv

from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Secrets import Secrets
from osbot_aws.apis.shell.Shell_Server import Shell_Server

from osbot_utils.utils.Misc import random_uuid

SHELL_VAR                = 'lambda_shell'
SHELL_VAR_AUTH_KEY       = 'auth_key'
SHELL_VAR_METHOD_NAME    = 'method_name'
SHELL_VAR_METHOD_KWARDS  = 'method_kwargs'
SHELL_VAR_METHOD_INVOKED = 'method_invoked'
SHELL_VAR_RETURN_VALUE   = 'return_value'
SHELL__ENV_VAR__AUTH_KEY = 'LAMBDA_SHELL__AUTH_KEY'

def add_lambda_shell_decorator(function_code):
    updated_code =  "from osbot_aws.apis.shell.Lambda_Shell import lambda_shell\n" +  \
                    "@lambda_shell\n" +                                               \
                    function_code
    return updated_code

# decorator
def lambda_shell(function):
    @wraps(function)
    def wrapper(event, context=None):
        shell_server = Lambda_Shell(event.get(SHELL_VAR))
        if shell_server.valid_shell_request():
            return shell_server.invoke()
        return function(event, context)
    return wrapper

class Lambda_Shell:
    def __init__(self, shell_command=None):
        self.secret_id     = 'lambda_shell_auth'
        self.secret        = Secrets(self.secret_id)
        self.shell_command = shell_command or {}


    @cache_on_self                                              # cache this value since it doesn't change in the lifetime of a lambda function
    def get_lambda_shell_auth(self):                            # this method can be used on both client and server
        data = getenv(SHELL__ENV_VAR__AUTH_KEY)
        if data:
            return data

        data = self.secret.value_from_json_string()             # this will create a new secret if it doesn't exist
        if data:
            return data.get('key')

    def reset_lambda_shell_auth(self):
        value = {'key' : random_uuid() }
        return Secrets(self.secret_id).update_to_json_string(value)

    def valid_shell_request(self):                             # also checks the AUTH
        if self.shell_command:
            if self.shell_command.get(SHELL_VAR_METHOD_NAME):
                expected_auth_key = self.get_lambda_shell_auth()
                request_auth_key = self.shell_command.get(SHELL_VAR_AUTH_KEY)
                if expected_auth_key and request_auth_key and expected_auth_key == request_auth_key:
                    return True
        return False

    def invoke(self):
        exec_result = Shell_Server().invoke(self.shell_command)
        if exec_result and exec_result.get(SHELL_VAR_METHOD_INVOKED):
            return exec_result.get(SHELL_VAR_RETURN_VALUE)


import json

import botocore
from osbot_utils.utils.Lists import unique

from osbot_utils.utils.Json import json_loads, json_parse

from osbot_utils.decorators.lists.group_by import group_by

from osbot_utils.testing.Duration import Duration

from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Logs                                import Logs
from osbot_utils.decorators.lists.index_by              import index_by
from osbot_utils.decorators.methods.catch               import catch
from osbot_utils.decorators.methods.remove_return_value import remove_return_value
from osbot_aws.apis.Session                             import Session
from osbot_aws.aws.s3.S3                                import S3
from osbot_utils.utils.Misc                             import wait, random_string, base64_to_str, list_set, wait_for
from osbot_utils.utils.Status                           import status_ok, status_error, status_warning


class Lambda:
    def __init__(self, name=''):
        self.architecture   = 'x86_64'                      # other option is 'arm64'
        self.runtime        = 'python3.11'
        self.memory         = 512                               # default to 512Mb max is 10,000 Mb (current AWS limit in Dec 2020)
        self.timeout        = 60                                # default to 60 secs (1m) max is 900 secs (15m)
        self.trace_mode     = 'PassThrough'                     # x-rays disabled
        self.original_name  = name
        self.handler        = name + '.run'
        self.name           = name.replace('.','_')
        self.folder_code    = None
        self.image_uri      = None                              # when using container images
        self.role           = None
        self.s3_bucket      = None
        self.s3_key         = None
        self.layers         = None
        self.env_variables  = None
        self.tags           = {}

    def __enter__(self): return self
    def __exit__ (self, exc_type, exc_val, exc_tb): pass

    # cached dependencies

    @cache_on_self
    def client(self):
        return Session().client('lambda')

    @cache_on_self
    def s3(self):
        return S3()

    # helper methods

    #@staticmethod
    # def invoke_using_paginator(api, method, field_id, **kwargs):
    #     paginator = api.get_paginator(method)
    #     for page in paginator.paginate(**kwargs):
    #         for id in page.get(field_id):
    #             yield id

    def _call_method_with_paginator(self, method, field_id, **kwargs):
        api       = self.client()
        paginator = api.get_paginator(method)
        for page in paginator.paginate(**kwargs):
            for id in page.get(field_id):
                yield id

    # main methods

    def add_layer(self, layer_arn=None):
        if layer_arn:
            if self.layers is None:
                self.layers = []
            self.layers.append(layer_arn)
            self.layers = unique(self.layers)           # make sure there are no duplicates
        return self

    def add_layers(self, layers_arn):
        for layer_arn in layers_arn:
            self.add_layer(layer_arn)
        return self

    def account_settings(self):
        return self.client().get_account_settings()

    @catch
    @remove_return_value(field_name='ResponseMetadata')
    def alias(self, function_name, name):
        return self.client().get_alias(FunctionName=function_name, Name=name)

    @catch
    def alias_create(self, function_name, function_version, name, description):
        return self.client().create_alias(FunctionName=function_name,FunctionVersion= function_version,Name=name, Description=description)

    @catch
    @remove_return_value(field_name='ResponseMetadata')
    def alias_delete(self, function_name, name):
        return self.client().delete_alias(FunctionName=function_name, Name=name)

    @index_by
    def aliases(self,function_name):
        return self.client().list_aliases(FunctionName=function_name).get('Aliases')

    @remove_return_value('ResponseMetadata')
    def configuration(self):
        return self.client().get_function_configuration(FunctionName=self.name)

    def configuration_update(self,**kvargs):
        return self.client().update_function_configuration(FunctionName=self.name,**kvargs)

    def create(self):
        kwargs = self.create_kwargs()
        kwargs_status = self.validate_create_kwargs(kwargs)
        if kwargs_status.get('status') != 'ok':
            return kwargs_status
        try:
            data = self.client().create_function(**kwargs)
            return { 'status': 'ok', 'name': self.name , 'data' : data }
        except Exception as error:
            return {'status': 'error', 'error': '{0}'.format(error), 'data': kwargs}

    def create_kwargs(self):
        kwargs = {  'Architectures' : [self.architecture]                                , # no idea why this is an array since we get an exception when there is more than one value
                    'FunctionName'  : self.name or  random_string(prefix='temp_lambda_') ,
                    'MemorySize'    : self.memory                                        ,
                    'Role'          : self.role                                          ,
                    'Timeout'       : self.timeout                                       ,
                    'TracingConfig' : { 'Mode': self.trace_mode }                        ,
                    'Tags'          : self.tags                                          }

        if self.env_variables:
            kwargs['Environment'] = {'Variables': self.env_variables}
        if self.image_uri:
            kwargs['Code'         ] = {'ImageUri': self.image_uri}
            kwargs['PackageType'  ] = "Image"
        else:
            kwargs['Code'          ] = {"S3Bucket": self.s3_bucket, "S3Key": self.s3_key}
            kwargs['PackageType'   ] = "Zip"
            kwargs['Runtime'       ] = self.runtime
            kwargs['Handler'       ] = self.handler

        if self.layers:
            kwargs['Layers'] = self.layers

        return kwargs


    def delete(self, delete_log_group=True):                # todo  Delete Lambda function method should also delete cloud formation logs #6 (https://github.com/owasp-sbot/OSBot-AWS/issues/6)
        if self.exists() is False:
            return False
        self.client().delete_function(FunctionName=self.name)
        if delete_log_group:
            self.log_group().group_delete()
        return self.exists() is False

    @catch
    def event_source_create(self, event_source_arn, function_name):
        return self.client().create_event_source_mapping(EventSourceArn=event_source_arn, FunctionName=function_name)

    @catch
    def event_source_delete(self, event_source_uuid):
        return self.client().delete_event_source_mapping(UUID=event_source_uuid)

    @catch
    @index_by
    def event_sources(self):
        return self.client().list_event_source_mappings().get('EventSourceMappings')

    def exists(self, function_name=None):
        try:
            return self.info(function_name or self.name) is not None
        except:
            return False

    @cache_on_self
    def function_arn(self):
        return self.info().get('Configuration').get('FunctionArn')

    def function_url(self, function_alias=None):
        return self.function_url_info(function_alias).get('FunctionUrl')

    def function_set_policy_to_allow_public_access(self):
        function_arn = self.function_arn()
        auth_type    = 'NONE'
        statement_id = 'FunctionURLAllowPublicAccess'
        if statement_id in self.policy_statements(index_by='Sid') :
            self.permission_delete(function_arn, statement_id)

        kwargs = { 'statement_id'          : statement_id                  ,
                   'action'                : 'lambda:InvokeFunctionUrl'    ,
                   'function_url_auth_type': auth_type                     ,
                   'function_arn'          : function_arn                  ,
                   'principal'             : '*'                           }

        return self.permission_add(**kwargs)

    def function_url_exists(self, function_alias=None):
        return self.function_url_info(function_alias) != {}

    def function_url_info(self, function_alias=None):
        try:
            kwargs = dict(FunctionName= self.function_arn())
            if function_alias:
                kwargs['Qualifier'] = function_alias

            return self.client().get_function_url_config(**kwargs)
        except Exception:                                  # todo: refactor to confirm exception is botocore.errorfactory.ResourceNotFoundException (which is not exposed)
            return {}

    def function_url_delete(self, function_alias=None):
        if self.function_url_exists(function_alias):
            kwargs = dict(FunctionName= self.function_arn())
            if function_alias:
                kwargs['Qualifier'] = function_alias

            return self.client().delete_function_url_config(**kwargs)

    @remove_return_value('ResponseMetadata')
    def function_url_create(self, function_alias=None, auth_type="AWS_IAM", cors=None, invoke_mode='BUFFERED'):
        if self.function_url_exists(function_alias) is False:
            function_arn   = self.function_arn()
            auth_type      = auth_type #'NONE',
            cors           = cors or {}
            invoke_mode    = invoke_mode # 'RESPONSE_STREAM'
            kwargs         = dict(FunctionName = function_arn   ,
                                  AuthType     = auth_type      ,
                                  Cors         = cors           ,
                                  InvokeMode   = invoke_mode    )
            if function_alias:
                kwargs['Qualifier'] = function_alias

            return self.client().create_function_url_config(**kwargs)
        return {"status": "warning", "message": "function url already existed"}

    def function_url_create_with_public_access(self, invoke_mode='BUFFERED'):
        function_url_create     = self.function_url_create(auth_type='NONE', invoke_mode=invoke_mode)
        self.wait_for_function_update_to_complete()
        function_set_policy     = self.function_set_policy_to_allow_public_access()
        self.wait_for_function_update_to_complete()
        return dict(function_url_create=function_url_create, function_set_policy=function_set_policy)

    def function_url_update(self, function_alias=None, auth_type=None, cors=None, invoke_mode=None):
        if self.function_url_exists(function_alias) :
            function_arn   = self.function_arn()

            kwargs         = dict(FunctionName = function_arn)

            if function_alias:
                kwargs['Qualifier'] = function_alias
            if auth_type:
                kwargs['AuthType'] = auth_type
            if cors:
                kwargs['Cors'] = cors
            if invoke_mode:
                kwargs['InvokeMode'] = invoke_mode

            return self.client().update_function_url_config(**kwargs)

    # todo: add this policy when creating a function url with NONE auth_type

    # {
    #   "Version": "2012-10-17",
    #   "Statement": [
    #     {
    #       "StatementId": "FunctionURLAllowPublicAccess",
    #       "Effect": "Allow",
    #       "Principal": "*",
    #       "Action": "lambda:InvokeFunctionUrl",
    #       "Resource": "{function_arn}",
    #       "Condition": {
    #         "StringEquals": {
    #           "lambda:FunctionUrlAuthType": "NONE"
    #         }
    #       }
    #     }
    #   ]
    # }
    @remove_return_value('ResponseMetadata')
    def info(self, function_name=None):
        return self.client().get_function(FunctionName=function_name or self.name)

    def invoke_raw(self, payload = None, client_context=None, log_type='None', qualifier=None):
        try:
            kwargs = { "FunctionName" : self.name                 ,
                       "Payload"      : json.dumps(payload or {}) ,
                       'LogType'      : log_type                  }
            if client_context:
                kwargs['ClientContext'] = client_context
            if qualifier:
                kwargs['Qualifier'    ] = qualifier

            response      = self.client().invoke(**kwargs)

            result_bytes  = response.get('Payload').read()
            result_string = result_bytes.decode('utf-8')
            result        = json_loads(result_string)
            if result == {}:
                result = {'raw_message': result_string}
            return { 'status': 'ok'   , 'name': self.name, 'data' : result , 'response': response }
        except Exception as error:
            return { 'status': 'error', 'name': self.name, 'data' : '{0}'.format(error) }

    def invoke(self, payload = None):
        result = self.invoke_raw(payload)
        if result.get('status') == 'ok':
            return result.get('data')
        return {'error': result.get('data')}

    def invoke_async(self, payload = None):
        kwargs = { "FunctionName"  : self.name                  ,
                   "Payload"       : json.dumps(payload or {})  ,
                   "InvocationType": 'Event'                    }
        return self.client().invoke(**kwargs)

    def invoke_dry_run(self, payload = None):
        kwargs = { "FunctionName"  : self.name                  ,
                   "Payload"       : json.dumps(payload or {})  ,
                   "InvocationType": 'DryRun'                   }
        return self.client().invoke(**kwargs)

    def invoke_return_logs(self, payload=None):
        result                   = self.invoke_raw(payload=payload, log_type='Tail')
        if result.get('status') == 'error':
            return result
        logs                     = result.get('response', {}).get('LogResult', '')
        result['execution_logs'] = base64_to_str(logs, encoding='utf-8')
        result['return_value'  ] = result.get('data')
        result['request_id'    ] = result.get('response', {}).get('ResponseMetadata',{}).get('RequestId')
        del result['data'    ]
        del result['response']
        return result

    @index_by
    def functions(self):
        return list(self._call_method_with_paginator('list_functions', 'Functions'))

    @index_by
    def functions_names(self):
        return list_set(self.functions(index_by='FunctionName'))

    def log_group(self):
        return Logs(group_name=f'/aws/lambda/{self.name}')

    def permission_add(self, function_arn, statement_id, action, principal, source_arn=None, function_url_auth_type=None):
        try:
            params = {  'FunctionName': function_arn,
                        'StatementId' : statement_id,
                        'Action'      : action,
                        'Principal'   : principal }
            if source_arn             : params['SourceArn'          ] = source_arn
            if function_url_auth_type : params['FunctionUrlAuthType'] = function_url_auth_type

            result = self.client().add_permission(**params)
            if type(result.get('Statement')) is str:
                return json_parse(result.get('Statement'))
            return result
        except Exception as error:
            return {'error': f'{error}'}

    def permission_delete(self, function_name, statement_id):
        try:
            params = {  'FunctionName': function_name,
                        'StatementId' : statement_id,                        }
            return self.client().remove_permission(**params)
        except Exception as error:
            return {'error': f'{error}'}

    def permissions(self):
        return self.policy().get('Statement',[])


    def shell(self):
        from osbot_aws.apis.shell.Shell_Client import Shell_Client                  # todo: see way to solve circular dependency
        return Shell_Client(self)

    def shell_python_exec(self, code, auth_key):  # this needs @lambda_shell on the Lambda's run method
        params = {'lambda_shell': {'method_name':'python_exec', 'method_kwargs': {'code' : code} ,'auth_key' : auth_key }}
        return self.invoke(params)

    @catch
    def policy(self):
        try:
            policy_str = self.client().get_policy(FunctionName=self.name).get('Policy')
            return json.loads(policy_str)
        except:                 # ResourceNotFoundException doesn't seem to exposed to we have to do a global capture
            return {}

    @index_by
    @group_by
    def policy_statements(self):
        return self.policy().get('Statement', [])

    def policy_statements_clear(self):
        function_arn = self.function_arn()
        for statement in self.policy_statements():
            statement_id = statement.get('Sid')
            self.permission_delete(function_name=function_arn, statement_id=statement_id)
        return self

    def set_handler             (self, value): self.handler        = value    ; return self
    def set_role                (self, value): self.role           = value    ; return self
    def set_s3_bucket           (self, value): self.s3_bucket      = value    ; return self
    def set_s3_key              (self, value): self.s3_key         = value    ; return self
    def set_folder_code         (self, value): self.folder_code    = value    ; return self
    def set_trace_mode          (self, value): self.trace_mode     = value    ; return self
    def set_layers              (self, value): self.layers         = value    ; return self
    def set_env_variables       (self, value): self.env_variables  = value    ; return self
#    def set_xrays_on            (self       ): self.trace_mode  = 'Active' ; return self

    # def set_s3_bucket_and_key(self, s3_bucket, s3_key):
    #     self.set_s3_bucket(s3_bucket)
    #     self.set_s3_key   (s3_key   )
    #     return self

    def set_env_variable(self, key, value):
        if self.env_variables is None:
            self.env_variables = {}
        self.env_variables[key] = value
        return self

    def upload(self):
        if self.image_uri is None:              # only upload when not using image_uri                                              # don't need to upload when running lambda from a container
            self.s3().folder_upload(self.folder_code, self.s3_bucket, self.s3_key)
            return self.s3().file_exists(self.s3_bucket, self.s3_key)

    def update(self):
        if self.exists() is False:
            self.upload()
            return self.create()
        else:
            self.upload()
        try:
            update_code_result   = self.update_lambda_code()
            update_config_result = self.update_lambda_configuration()
            return {'status': 'ok'    , 'name': self.name, 'data': {"update_code": update_code_result, "update_config": update_config_result}}

        except Exception as error:
            return { 'status': 'error', 'name': self.name, 'data': '{0}'.format(error)}

    def update_lambda_code(self):
        self.wait_for_function_update_to_complete()
        if self.image_uri:
            update_kwargs = dict(FunctionName = self.name     ,
                                 ImageUri     = self.image_uri)
        else:
            update_kwargs = dict(FunctionName = self.name     ,
                                 S3Bucket     = self.s3_bucket,
                                 S3Key        = self.s3_key   )

        return self.client().update_function_code(**update_kwargs)

    def update_lambda_image_uri(self, image_uri):
        self.wait_for_function_update_to_complete()
        self.image_uri = image_uri
        return self.client().update_function_code(FunctionName = self.name     ,
                                                  ImageUri     = self.image_uri)

    def update_lambda_configuration(self):
        self.wait_for_function_update_to_complete()         # make sure there is no update also happening at this time
        kwargs = {}
        if self.layers:
            kwargs['Layers'] = self.layers
        if self.env_variables:
            kwargs['Environment'] = {'Variables': self.env_variables}
        if kwargs != {}:
            kwargs['FunctionName'] = self.name
            return self.client().update_function_configuration(**kwargs)

    def validate_create_kwargs(self, kwargs):
        name      = kwargs.get('FunctionName'    )
        code      = kwargs.get('Code'        , {})
        image_uri = code  .get('ImageUri'        )
        s3_bucket = code  .get('S3Bucket'        )
        s3_key    = code  .get('S3Key'           )

        if len(name) > 64:
            return status_error(message="lambda functions name cannot be bigger than 64 chars", data=self.name)
        if self.exists(function_name=name):
            return status_warning(message=f'lambda function already exists: {name}')

        if image_uri is None:
            if self.s3().file_not_exists(s3_bucket, s3_key):
                return status_error(message=f'for function {name}, could not find provided s3 bucket and s3 key: {s3_bucket} {s3_key}')
        return status_ok(message='validated ok  the lambda create_kwargs')

    def wait_for_function_update_to_complete(self, max_attempts=40, wait_time=0.1):
        status = None
        for i in range(max_attempts):
            configuration = self.configuration()
            status        = configuration.get('LastUpdateStatus')
            if status == 'Successful':
                #print(f"\n after  {i} attempts, update completed successfully")
                break
            status = configuration.get('State')         # helps to debug when the update fails
            wait_for(wait_time)
        return status

    def wait_for_state(self, state, max_wait_count=40, wait_interval=1):
        for i in range(0, max_wait_count):
            info = self.info()
            current_state = info.get('Configuration').get('State')
            if current_state == state:
                return {"status" : "ok", "message": f"Status '{state}' was found after {i} * {wait_interval} seconds"}
            wait(wait_interval)
        return {"status" : "error", "message": f"Status '{state}' did not occur in {i} * {wait_interval} seconds"}

    def wait_for_state_active(self, **kwargs):
        return self.wait_for_state("Active", **kwargs)

def invoke_lambda(lambda_name, payload):
    return Lambda(name=lambda_name).invoke(payload)

def invoke_lambda_async(lambda_name, payload):
    return Lambda(name=lambda_name).invoke_async(payload)



from osbot_aws.aws.dynamo_db.models.DyDB__Document import DyDB__Document


class DyDB__Resource(DyDB__Document):               # todo: figure out if we need this helper class
    pass                                            #       which would have common methods to all Resource classes like Resource__Config



class Session_Client_Creation_Fail(Exception):

    def __init__(self, status):
        self.status  = status
        self.message = status.get('message')
        self.data    = status.get('data')
        self.error   = status.get('error')
        super().__init__(self.message)

    def __str__(self):
        return self.message


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

# for details on this model see https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-titan-text.html
# 'amazon.titan-text-express-v1'    # working on this
# 'amazon.titan-text-lite-v1'   ,
# 'amazon.titan-tg1-large'
class Amazon_Titan__Text(Kwargs_To_Self):
    # prompt vars
    input_text     : str    = ""
    temperature    : float  = 0.0
    topP           : int    = 1
    max_token_count: int    = 1024
    stop_sequences : list

    def body(self, prompt=None):
        if prompt:
            self.input_text = prompt
        return   { "inputText"           : self.input_text,
                   "textGenerationConfig": { "temperature"  : self.temperature     ,
                                             "topP"         : self.topP            ,
                                             "maxTokenCount": self.max_token_count ,
                                             "stopSequences": self.stop_sequences  }}


from osbot_aws.apis.Events import Events
from osbot_utils.utils.Json import json_to_str
from osbot_utils.utils.Misc import random_string, date_time_now


class Event_Rule:
    def __init__(self, rule_name=None, event_source=None, tags=None, event_bus_name=None):
        self.event_bus_name = event_bus_name
        self.event_source   = event_source or random_string(prefix='temp_event_source-')
        self.rule_name      = rule_name    or random_string(prefix='temp_event_rule-')
        self.rule_tags      = tags         or {'Name': self.event_source}
        self.events         = Events()

    def add_target_sqs_queue(self, sqs_queue):
        rule              = self.info()
        rule_name         = self.rule_name
        target_id         = sqs_queue.name()
        target_arn        = sqs_queue.arn()
        source_arn        = rule.get('Arn')
        service           = 'events.amazonaws.com'
        resource          = sqs_queue.arn()
        target_attributes = { 'SqsParameters': { 'MessageGroupId': rule_name}}
        sqs_queue.permission_add_for_service(source_arn=source_arn, service=service, resource=resource)
        self.events.target_create(rule_name=rule_name, target_id=target_id, target_arn=target_arn, target_attributes=target_attributes)
        return target_id

    def create(self):
        return self.events.rule_create(rule_name=self.rule_name, event_source=self.event_source, tags=self.rule_tags, event_bus_name=self.event_bus_name)

    def delete(self):
        return self.events.rule_delete(self.rule_name, event_bus_name=self.event_bus_name)

    def delete_target(self, target_id=None):
        return self.events.target_delete(rule_name=self.rule_name, target_id=target_id)

    def exists(self):
        return self.events.rule_exists(rule_name=self.rule_name,event_bus_name=self.event_bus_name)

    def info(self):
        return self.events.rule(rule_name=self.rule_name, event_bus_name=self.event_bus_name)

    def name(self):
        return self.rule_name

    def send_event(self, event_data):
        if type(event_data) is not str:
            event_data = json_to_str(event_data, pretty=False)
        event  = {   'Time'      : date_time_now()       ,
                     'Source'    : self.event_source,
                     "DetailType": "myTestType"     ,
                     'Detail'    : event_data       }
        return self.events.events_put([event])

    def targets(self):
        return self.events.targets(rule_name=self.rule_name, event_bus_name=self.event_bus_name)


from unittest import TestCase

import pytest

from osbot_aws.aws.bedrock.Bedrock                  import Bedrock
from osbot_aws.aws.bedrock.Bedrock__with_temp_role  import Bedrock__with_temp_role
from osbot_aws.aws.bedrock.cache.Bedrock__Cache     import Bedrock__Cache
from osbot_aws.testing.Pytest                       import skip_pytest___aws_pytest_user_name__is_not_set
from osbot_utils.utils.Env                          import in_github_action


class TestCase__Bedrock(TestCase):
    bedrock     : Bedrock
    region_name : str
    cache       : Bedrock__Cache


    @classmethod
    def setUpClass(cls):
        skip_pytest___aws_pytest_user_name__is_not_set()
        if in_github_action():              # skipping all tests in GH actions, since there is no cache in there are we are just having cost with no value                             # disabling Bedrock tests in GitHub actions since they are not 100% deterministic
            pytest.skip()
        cls.region_name = 'us-east-1'
        cls.bedrock = Bedrock__with_temp_role(region_name=cls.region_name)
        cls.cache   = cls.bedrock.bedrock_cache

    @classmethod
    def tearDownClass(cls):
        pass



import sys


class Session_No_Credentials(Exception):

    def __init__(self, print_error_message= True):
        message = (f"\tNo Security credentials where found\n\n"
                   f"\tPlease check: current environment values, "
                   f"profile name or value in your ~/.aws/credentials file")
        self.print_error_message = print_error_message
        if self.print_error_message:
            print()
            print("\n*********** OSBot_AWS ***********\n")
            print(f"{message}")
            print("\n*********** OSBot_AWS ***********\n")

        super().__init__(message)


from functools import cache

from osbot_utils.decorators.methods.remove_return_value import remove_return_value

#from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.aws.iam.IAM import IAM

class IAM_User:
    def __init__(self, user_name=None):
        self.user_name = user_name

    @cache
    def iam(self):
        return IAM(user_name=self.user_name)

    @cache
    def user(self):
        return self.iam().resource().User(name=self.user_name)

    def arn(self):
        return self.user().arn

    def create(self):
        return self.iam().user_create()

    def delete(self):
        return self.iam().user_delete()

    def exists(self):
        return self.iam().user_exists()

    def _get_attribs(self,target,attribs):
        item = {}
        for attrib in attribs:
            item[attrib] = getattr(target, attrib)
        return item

    def _index_by(self, target, attribs, index_by=None):
        try:
            all_data = getattr(target,'all')()
            results  = []
            for value in all_data:
                results.append(self._get_attribs(value,attribs))
            if index_by is None:
                return results
            indexed = {}
            for item in results:
                indexed[item.get(index_by)] = item
            return indexed
        except Exception as error:
            return {'error': f'{error}'}

    def access_keys(self,index_by=None):
        attribs = ['user_name','id','access_key_id', 'create_date','status']
        return self._index_by(self.user().access_keys, attribs, index_by)

    def groups(self, index_by=None):
        return self._index_by(self.user().groups, ['arn','create_date','group_id','group_name','path'], index_by)

    def user_id(self):
        return self.user().user_id

    def info(self):
        attribs = ['arn', 'create_date', 'password_last_used', 'path', 'permissions_boundary', 'tags', 'user_id', 'user_name']
        return self._get_attribs(self.user(), attribs)

    def policies(self, index_by=None):
        attribs = ['attachment_count', 'create_date', 'default_version_id', 'description', 'is_attachable', 'path', 'permissions_boundary_usage_count', 'policy_id', 'policy_name', 'update_date']
        return self._index_by(self.user().attached_policies,attribs, index_by)

    # def roles(self):
    #     return self.iam().roles()

    @remove_return_value('ResponseMetadata')
    def set_password(self, password, reset_required=True):
        try:
            return self.iam().login_profile_create(password, reset_required=reset_required)
        except Exception as error:
            return {'error' : f'{error}'}

    def user_policies(self, index_by=None):
        attribs = ['attachment_count', 'create_date', 'default_version_id', 'description', 'is_attachable', 'path', 'permissions_boundary_usage_count', 'policy_id', 'policy_name', 'update_date']
        return self._index_by(self.user().policies,attribs, index_by)

    def tags(self):
        return self.user().tags


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self

# for details see  https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-claude.html

class Anthropic__Claude_Instant_V1(Kwargs_To_Self):
    # config vars
    model_id            : str = 'anthropic.claude-instant-v1'
    prompt_format       : str = "\n\nHuman:{prompt}\n\nAssistant:"
    # prompt vars
    prompt              : str
    temperature         : float = 0.0
    top_p               : int   = 1
    top_k               : int   = 0
    max_tokens_to_sample: int   = 1024
    stop_sequences      : list

    def body(self):
        anthropic_prompt = self.prompt_format.format(prompt=self.prompt)
        return  { "prompt"               : anthropic_prompt          ,
                  "temperature"          : self.temperature          ,
                  "top_p"                : self.top_p                ,
                  "top_k"                : self.top_k                ,
                  "max_tokens_to_sample" : self.max_tokens_to_sample ,
                  "stop_sequences"       : self.stop_sequences       }


from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self


class WS__Handle_Lambda(Kwargs_To_Self):

    def handle_lambda(self, event, context):
        connection_id = event.get('requestContext', {}).get('connectionId', '')
        route_key     = event.get('requestContext', {}).get('routeKey', '')

        if route_key == '$connect':                                     # Handle the connection event
            return self.route__connect(event, connection_id)

        if route_key == '$disconnect':                                  # Handle the disconnection event
            return self.route_disconnect(event, connection_id)

        if route_key == '$default':                                     # Handle the default route
            return self.route_default(event, connection_id)

        return self.route_unknown(event, connection_id, route_key)      # Handle an unknown route

    def route__connect(self, event, connection_id):
        return {'statusCode': 200, 'body': 'Connected.'                         }

    def route_disconnect(self, event, connection_id):
        return {'statusCode': 200, 'body': 'Disconnected.'                      }

    def route_default(self, event, connection_id):
        return {'statusCode': 200, 'body': 'Message received on default route.' }

    def route_unknown(self, event, connection_id, route_key):
        return {'statusCode': 400, 'body': 'Route not recognized.'              }


from osbot_aws.aws.dynamo_db.domains.DyDB__Table_With_GSI import DyDB__Table_With_GSI
from osbot_utils.utils.Misc import timestamp_utc_now, timestamp_utc_now_less_delta

TIMESTAMP_SORT_KEY          = 'timestamp'
TIMESTAMP_SORT_KEY_TYPE     = 'N'
TIMESTAMP_SORT_KEY_SCHEMA   = 'RANGE'
TIMESTAMP_PROJECTION_TYPE   = 'ALL'
INDEX_PARTITION_TYPE        = 'S'

class DyDB__Table_With_Timestamp(DyDB__Table_With_GSI):

    table_indexes : list

    def __init__(self,**kwargs):
        super().__init__(**kwargs)

    def add_document(self,document):
        timestamp = document.get(TIMESTAMP_SORT_KEY)
        if not timestamp:
            document[TIMESTAMP_SORT_KEY] = timestamp_utc_now()
        return super().add_document(document)


    def index_create_kwargs(self, index_name):                   # every index here will have a timestamp has default
        kwargs = dict(partition_name  = index_name                ,
                      partition_type  = INDEX_PARTITION_TYPE      ,
                      sort_key        = TIMESTAMP_SORT_KEY        ,
                      sort_key_type   = TIMESTAMP_SORT_KEY_TYPE   ,
                      sort_key_schema = TIMESTAMP_SORT_KEY_SCHEMA ,
                      projection_type = TIMESTAMP_PROJECTION_TYPE )
        index_create_kwargs = self.gsi_create_kwargs(**kwargs).get('data')
        return index_create_kwargs

    def indexes_create_kwargs(self):
        all_attribute_definitions = []
        all_gsi_updates           = []
        indexes_create_kwargs = {'attribute_definitions': all_attribute_definitions ,
                                 'gsi_updates'          : all_gsi_updates           }
        if self.table_indexes:
            for index_name in self.table_indexes:
                index_create_kwargs = self.index_create_kwargs(index_name)
                attribute_definitions = index_create_kwargs.get('attribute_definitions')
                gsi_update            = index_create_kwargs.get('gsi_update'           )
                for attribute_definition in attribute_definitions:
                    if attribute_definition not in all_attribute_definitions:
                        all_attribute_definitions.append(attribute_definition)
                all_gsi_updates.append(gsi_update)
        return indexes_create_kwargs

    def create_table_kwargs(self):
        create_table_kwargs               = super().create_table_kwargs()
        new_indexes_create_kwargs         = self.indexes_create_kwargs()
        new_indexes_attribute_definitions = new_indexes_create_kwargs.get('attribute_definitions')
        new_indexes_gsi_updates           = new_indexes_create_kwargs.get('gsi_updates')
        table_attribute_definitions       = create_table_kwargs.get('AttributeDefinitions')
        global_secondary_indexes          = []

        for attribute_definitions in new_indexes_attribute_definitions:
            if attribute_definitions not in table_attribute_definitions:
                table_attribute_definitions.append(attribute_definitions)
        for gsi_update in new_indexes_gsi_updates:
            create_index_config = gsi_update.get('Create')
            global_secondary_indexes.append(create_index_config)

        if global_secondary_indexes:
            create_table_kwargs['GlobalSecondaryIndexes'] = global_secondary_indexes

        return create_table_kwargs

    def query_index_last_n_hours(self, index_name, index_value, hours, query_filter=None):
        timestamp_start = timestamp_utc_now_less_delta(hours=hours)  # get timestamp for last n hours
        timestamp_end = timestamp_utc_now()
        return self.query_index_by_timestamp(index_name, index_value, timestamp_start, timestamp_end, query_filter)

    def query_index_by_timestamp(self, index_name, index_value, timestamp_start, timestamp_end, query_filter=None):
        query_kwargs = dict(index_name    = index_name              ,
                            index_type    = INDEX_PARTITION_TYPE    ,
                            index_value   = index_value             ,
                            sort_key      = TIMESTAMP_SORT_KEY      ,
                            sort_key_type = TIMESTAMP_SORT_KEY_TYPE ,
                            start_value   = timestamp_start         ,
                            end_value     = timestamp_end           ,
                            query_filter  = query_filter            )

        return self.query_index_between_range(**query_kwargs)


    def setup(self):
        with self as _:
            if _.not_exists():
                self.create_table(wait_for_table=False)




from boto3.dynamodb.types import TypeSerializer

from osbot_aws.aws.apigateway.apigw_dydb.WS__Connection         import WS__Connection
from osbot_aws.aws.dynamo_db.domains.DyDB__Table_With_Timestamp import DyDB__Table_With_Timestamp
from osbot_aws.utils.Execution_Env                              import current_execution_env
from osbot_utils.utils.Json                                     import to_json_str, from_json_str
from osbot_utils.utils.Misc                                     import date_time_now, timestamp_utc_now, timestamp_to_str_time

DYNAMO_DB__TABLE_NAME__WEB_SOCKETS = 'osbot__web_sockets'
TABLE_WEB_SOCKETS__INDEXES_NAMES   = ['date', 'status']

WS_CONNECTION_STATE__ACTIVE        = 'active'
WS_CONNECTION_STATE__DISCONNECTED  = 'disconnected'

class WS__DyDB(DyDB__Table_With_Timestamp):

    env: str = current_execution_env()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.table_name    = DYNAMO_DB__TABLE_NAME__WEB_SOCKETS
        self.table_indexes = TABLE_WEB_SOCKETS__INDEXES_NAMES

    def handle_route(self, connection_id, source=None, lambda_event=None):
        timestamp         = timestamp_utc_now()
        lambda_event_body = (lambda_event or {}).get('body')
        request_data      = from_json_str(lambda_event_body)
        if not request_data:
            request_data = {"request": lambda_event_body}

        response_message = self.handle_request_data(connection_id, request_data)

        topic = 'ws__dydb'
        data  = { 'action'        : 'show_message'  ,
                  'connection_id' : connection_id   ,
                  'response'      : response_message}
        #data.update(request_message)

        client_message = dict(when=timestamp_to_str_time(timestamp),
                              topic=topic,
                              data=data)

        body = to_json_str(client_message)

        return {'statusCode': 200, 'body': body }

    def register_connection(self, connection_id, source=None, lambda_event=None):
        status             = WS_CONNECTION_STATE__ACTIVE
        event_for_action   = { 'action'      : 'connect'           ,
                               'timestamp'   : timestamp_utc_now() ,
                               'source'      : source              ,
                               'status'      : status              ,
                               'lambda_event': lambda_event        }

        kwargs             = dict(id            = connection_id       ,  # index
                                  status        = status              ,  # index
                                  env           = self.env            ,
                                  events        = [event_for_action]  )

        ws_connection      = WS__Connection(**kwargs)
        document           = ws_connection.json()
        response = super().add_document(document)
        if response.get('document'):
            return response.get('document', {})
        return response

    def register_disconnect(self, connection_id, source=None, lambda_event=None):
        status             = WS_CONNECTION_STATE__DISCONNECTED
        timestamp          = timestamp_utc_now()
        event_for_action   = { 'action'      : 'disconnect'        ,
                               'timestamp'   : timestamp           ,
                               'source'      : source              ,
                               'status'      : status              ,
                               'lambda_event': lambda_event        }
        dydb_document = self.dydb_document(connection_id, load_data=False)
        dydb_document.add_to_list('events', event_for_action       )       # todo: add dydb_document ability to update different types of fields
        dydb_document.set_field  ('status', status                 )       #       like in this case: append to a list and set couple values
        dydb_document.set_field  ('timestamp_disconnect', timestamp)       # todo : add dydb_document.set_fields feature (to prevent multiple calls)


    def handle_request_data(self, connection_id, request_data):
        command = request_data.get('command')                                                # todo: see if should have this here, or if handling these commands should be done by the ones who implement this class
        if command == 'register':
            name  = request_data.get('name' , 'NA')
            state = request_data.get('state', 'NA')
            dydb_document = self.dydb_document(connection_id, load_data=False)
            dydb_document.set_field ('connection_name', name)                               # todo: replace with call to set multiple fields at the same time
            dydb_document.set_field('state'          , state)
            return {'message': f'registered device {name} with state {state}'}
        if command == 'Whoami':
            dydb_document = self.dydb_document(connection_id)
            connection_name = dydb_document.value('connection_name')
            state           = dydb_document.value('state')
            return f"you are '{connection_name}' and are in the state {state}"

        return {'message': f'no command provided : {request_data}' }


from osbot_utils.decorators.lists.index_by import index_by

from osbot_utils.decorators.lists.group_by import group_by

from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.Session import Session

class SSM:

    def __init__(self):
        pass

    @cache_on_self
    def client(self):
        return Session().client('ssm')

    def command_run(self, instance_id, command):
        resp = self.client().send_command(
            DocumentName="AWS-RunShellScript",
            Parameters={'commands': [command]},
            InstanceIds=[instance_id],
        )
        return resp

    def commands_list(self):
        return self.client().list_commands()

    @index_by
    @group_by
    def parameters(self, filter_value, filter_key='Name', filter_option='Equals'):
        kwargs = { "ParameterFilters": [{ 'Key'   : filter_key         ,
                                          'Option': filter_option      ,
                                          'Values': [filter_value]    }]}
        return self.client().describe_parameters(**kwargs).get("Parameters")


import uuid

from osbot_aws.aws.dynamo_db.Dynamo_DB                  import Dynamo_DB
from osbot_aws.aws.dynamo_db.Dynamo_DB__Record          import Dynamo_DB__Record
from osbot_utils.base_classes.Kwargs_To_Self            import Kwargs_To_Self
from osbot_utils.decorators.methods.capture_status      import apply_capture_status
from osbot_utils.utils.Misc                             import timestamp_utc_now, wait_for

PRIMARY_KEY_NAME = 'id'
PRIMARY_KEY_TYPE = 'S'

@apply_capture_status
class Dynamo_DB__Table(Kwargs_To_Self):
    dynamo_db  : Dynamo_DB
    key_name   : str
    key_type   : str
    table_name : str

    def __init__(self, **kwargs):
        self.key_name = PRIMARY_KEY_NAME
        self.key_type = PRIMARY_KEY_TYPE
        super().__init__(**kwargs)

    def add_document(self, document):
        key_value = document.get(self.key_name)
        if not key_value:                                                   # if key value is not set
            document = {**document ,                                        # to the current document object
                        self.key_name : self.dynamo_db.random_id()}         # add it as a random UUID
        return self.dynamo_db.document_add(table_name=self.table_name, document=document)

    def add_documents(self, documents):
        documents_to_add = []
        for document in documents:
            if self.key_name not in document:                               # If key is present,
                document = {self.key_name: self.dynamo_db.random_id(),      # add it as a generate a random UUID as the key
                            **document}                                     # to the current document object
            documents_to_add.append(document)
        return self.dynamo_db.documents_add(table_name=self.table_name, documents=documents_to_add)

    def add_record(self, record : Dynamo_DB__Record):
        metadata = record.metadata
        metadata.timestamp_created = timestamp_utc_now()
        document   = record.serialize_to_dict()
        add_result = self.add_document(document)
        if add_result.get('status') == 'ok':
            return add_result.get('data')
        raise Exception(f'Error adding record: {add_result}')

    def create_table(self):
        return self.dynamo_db.table_create(table_name=self.table_name, key_name=self.key_name)

    def clear_table(self):
        return self.dynamo_db.documents_delete_all(table_name=self.table_name, key_name=self.key_name)

    def delete_document(self, key_value):
        return self.dynamo_db.document_delete(table_name=self.table_name, key_name=self.key_name, key_value=key_value)

    def delete_documents(self, keys_values):
        return self.dynamo_db.documents_delete(table_name=self.table_name, key_name=self.key_name, keys_values=keys_values)

    def delete_table(self, wait_for_deletion=True):
        return self.dynamo_db.table_delete(table_name=self.table_name, wait_for_deletion=wait_for_deletion)

    def document(self, key_value):
        return self.dynamo_db.document(table_name=self.table_name, key_name=self.key_name, key_value=key_value)

    def documents(self, keys_values):
        return self.dynamo_db.documents(table_name=self.table_name, key_name=self.key_name, keys_values=keys_values)

    def documents_all(self):
        return self.dynamo_db.documents_all(table_name=self.table_name)

    def documents_ids(self,**kwargs):
        return self.dynamo_db.documents_ids(table_name=self.table_name, key_name=self.key_name, **kwargs)

    def exists(self):
        return self.dynamo_db.table_exists(table_name=self.table_name)

    def info(self):
        return self.dynamo_db.table_info(table_name=self.table_name)

    def keys(self):
        return self.dynamo_db.documents_keys(table_name=self.table_name, key_name=self.key_name, key_type=self.key_type)

    def query(self, **kwargs):
        return self.dynamo_db.client().query(**kwargs)          # todo: refactor this to add values that we know here (like TableName)

    def status(self):
        return self.dynamo_db.table_status(table_name=self.table_name)

    def update_table(self, attribute_definitions=None, gsi_updates=None):
        return self.dynamo_db.table_update(self.table_name, attribute_definitions=attribute_definitions, gsi_updates=gsi_updates)

    def gsi_create(self, partition_name, partition_type, sort_key=None, sort_key_type=None, sort_key_schema=None, projection_type='ALL'):
        gsi_create_kwargs= self.gsi_create_kwargs( partition_name  = partition_name    ,
                                                   partition_type  = partition_type    ,
                                                   sort_key        = sort_key          ,
                                                   sort_key_type   = sort_key_type     ,
                                                   sort_key_schema = sort_key_schema   ,
                                                   projection_type = projection_type   ).get('data')
        return self.gsi_create_index(**gsi_create_kwargs).get('data')               # todo: add check for status =='ok'

    def gsi_create_index(self, attribute_definitions, gsi_update):
        gsi_updates = [gsi_update]
        return self.gsi_create_indexes(attribute_definitions, gsi_updates)

    def gsi_create_indexes(self, attribute_definitions, gsi_updates):
        result = self.update_table( attribute_definitions=attribute_definitions, gsi_updates=gsi_updates)
        if result.get('status') == 'ok':
            return result.get('data')
        return result

    def gsi_create_kwargs(self, partition_name, partition_type, sort_key=None, sort_key_type=None, sort_key_schema=None, projection_type='ALL'):
        attribute_definitions = [{ 'AttributeName': partition_name,
                                   'AttributeType': partition_type}]
        gsi_update            = {'Create': { 'IndexName' : partition_name                        ,
                                             'KeySchema' : [{ 'AttributeName' : partition_name   ,
                                                              'KeyType'       : 'HASH'           }],
                                             'Projection':  { 'ProjectionType': projection_type  }}}
        if sort_key and sort_key_type and sort_key_schema:
            attribute_definitions.append({ 'AttributeName': sort_key,
                                           'AttributeType': sort_key_type})
            gsi_update.get('Create').get('KeySchema').append({ 'AttributeName' : sort_key           ,
                                                               'KeyType'       : sort_key_schema })
        create_kwargs = dict(attribute_definitions = attribute_definitions,
                             gsi_update            = gsi_update)
        return create_kwargs

    def gsi_delete(self, index_name):
        gsi_update  = {'Delete': { 'IndexName' : index_name }}
        gsi_updates = [gsi_update]
        return self.update_table(gsi_updates=gsi_updates).get('data')

    def gsi_wait_for_status(self, status='ACTIVE', max_attempts=20, delay=0.05):        # todo: see if these values need to be higher when dealing with AWS DynamoDB (vs the dynamodb-local)
        for i in range(max_attempts):
            all_match = False
            for gsi in self.gsis().get('data'):
                if gsi.get('IndexStatus') == status:
                    all_match = True
                else:
                    all_match = False
                    break

            if all_match:
                return True
            else:
                wait_for(delay)
        raise ValueError("expected status not found")


    def gsis(self):
        return self.info().get('GlobalSecondaryIndexes')


from osbot_aws.aws.iam.roles.IAM__Statement__Principal import IAM__Statement__Principal


class IAM__Policy__Assume__Service(IAM__Statement__Principal):
    action       : str =  'sts:AssumeRole'
    effect       : str = 'Allow'

    def __init__(self, service_name):
        self.principal_type  = 'Service'
        self.principal_value = f'{service_name}.amazonaws.com'
        super().__init__()


def iam_statement__codebuild__assume_service():
    return IAM__Policy__Assume__Service('codebuild').statement()

def iam_statement__ec2__assume_service():
    return IAM__Policy__Assume__Service('ec2').statement()

def iam_statement__lambda__assume_service():
    return IAM__Policy__Assume__Service('lambda').statement()


from osbot_aws.aws.iam.IAM_Policy                            import IAM_Policy
from osbot_aws.aws.iam.roles.IAM__Statement__Resource__Allow import IAM__Statement__Resource__Allow

class IAM__Policy__Service__Full_Access(IAM__Statement__Resource__Allow):
    policy_name : str

    def __init__(self, service_name):
        actions  = [f'{service_name}:*']
        resource = '*'
        self.policy_name = f"Policy__Full_Access__{service_name}"
        super().__init__(actions=actions, resource=resource)

    def arn(self):
        return self.iam_policy().policy_arn

    def iam_policy(self):
        return IAM_Policy(policy_name=self.policy_name, statements=[self.statement()])

    def create(self):
        return self.iam_policy().create()


from botocore.exceptions                    import ClientError

from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Lists import list_get

from osbot_aws.AWS_Config                   import AWS_Config
from osbot_utils.decorators.lists.group_by  import group_by
from osbot_aws.apis.Cloud_Watch_Logs        import Cloud_Watch_Logs
from osbot_utils.decorators.lists.index_by  import index_by
from osbot_utils.decorators.methods.cache   import cache
from osbot_aws.apis.Session                 import Session
from osbot_utils.utils.Misc                 import wait

from osbot_aws.aws.iam.IAM_Role import IAM_Role
from osbot_aws.aws.sts.STS import STS

# todo: find good solution to capture/manage these config values (See also values in EC2 class)
ECS_WAITER_DELAY        = 1             # default was 6 seconds
ECS_WAITER_MAX_ATTEMPTS = 600           # default was 100 times

class ECS:

    def __init__(self):
        self.aws_config = AWS_Config()
        self.account_id = self.aws_config.aws_session_account_id()
        self.region     = self.aws_config.aws_session_region_name()

    def __enter__(self                           ): return self
    def __exit__ (self, exc_type, exc_val, exc_tb): pass

    @cache
    def client(self):
        return Session().client('ecs')

    @cache
    def sts(self):
        return STS()

    @cache
    def cloud_watch_logs(self):
        return Cloud_Watch_Logs()

    @index_by
    def cluster(self, cluster_arn=None):
        if cluster_arn is None:
            cluster_arn = self.cluster_arn()
        result = self.clusters(clusters_arns=[cluster_arn])
        if len(result) == 1:
            return result[0]

    def cluster_arn(self, cluster_name='default'):
        return f'arn:aws:ecs:{self.region}:{self.account_id}:cluster/{cluster_name}'

    def cluster_create(self, cluster_name):
        cluster_arn = self.cluster_arn(cluster_name=cluster_name)
        if self.cluster_exists(cluster_arn):
            return self.cluster(cluster_arn)
        else:
            response = self.client().create_cluster(clusterName=cluster_name)
            return response.get('cluster')
        
    def cluster_delete(self, cluster_arn):
        response = self.client().delete_cluster(cluster=cluster_arn)
        return response.get('cluster')

    def cluster_exists(self, cluster_arn):
        cluster = self.cluster(cluster_arn=cluster_arn)
        if cluster and cluster.get('status') == 'ACTIVE':       # threat inactive cluster as non existent
            return True
        return False

    @index_by
    def clusters(self, clusters_arns=None):
        include         = [ 'ATTACHMENTS' , 'SETTINGS' , 'STATISTICS' , 'TAGS']
        target_clusters = clusters_arns or self.clusters_arns()
        return self.client().describe_clusters(clusters=target_clusters, include=include).get('clusters')

    def clusters_arns(self):
        return self.client().list_clusters().get('clusterArns')

    def container(self, task_arn, image_name, cluster_name='default'):
        return self.containers(task_arn=task_arn, cluster_name=cluster_name, index_by='name').get(image_name)

    def container_definition(self, task_definition_arn, image_name):
       return self.containers_definitions(task_definition_arn=task_definition_arn, index_by='name').get(image_name)

    @index_by
    def containers(self, task_arn, cluster_name='default'):
        return self.task(cluster_name=cluster_name, task_arn=task_arn).get('containers')

    @index_by
    def containers_definitions(self,task_definition_arn):
        return self.task_definition(task_definition_arn=task_definition_arn).get('containerDefinitions')

    @index_by
    @group_by
    def container_instances(self, cluster_name='default', instance_status='ACTIVE'):
        container_instances_arns = self.container_instances_arns(cluster_name=cluster_name, instance_status=instance_status)
        kwargs = { "cluster"            : cluster_name             ,
                   "containerInstances" : container_instances_arns }

        container_instances = self.client().describe_container_instances(**kwargs).get('containerInstances')
        # fix container_instance.get('attributes') since they are not in easy to consume key:value pairs
        for container_instance in container_instances:
            attributes_original = container_instance.get('attributes')
            attributes          = {}
            for attribute in attributes_original:
                key   = attribute.get('name')
                value = attribute.get('value')
                attributes[key] = value
            container_instance['attributes'] = attributes

        return container_instances

    def container_instances_arns(self, cluster_name='default', instance_status='ACTIVE'):
        kwargs = { "cluster": cluster_name }
        if instance_status:
            kwargs["status"] = instance_status

        return self.client().list_container_instances(**kwargs).get('containerInstanceArns')

    def log_stream_config(self, task_arn, task_definition_arn, image_name, cluster_name='default'):
        task                 = self.task(cluster_name=cluster_name, task_arn=task_arn)
        task_arn             = task.get('taskArn')
        task_id              = task_arn.split('/').pop()
        container_definition = self.container_definition(task_definition_arn=task_definition_arn, image_name=image_name)
        log_configuration    = container_definition.get('logConfiguration')
        #image_name           = container_definition.get('name')
        log_group_name       = log_configuration.get('options').get('awslogs-group')
        log_stream_prefix    = log_configuration.get('options').get('awslogs-stream-prefix')
        log_stream_name      = f'{log_stream_prefix}/{image_name}/{task_id}'
        config               = {"log_group_name" : log_group_name  ,
                                "log_stream_name": log_stream_name }
        return config

    def log_stream_exists(self, task_arn, task_definition_arn, image_name, cluster_name='default'):
        config          = self.log_stream_config(task_arn=task_arn, task_definition_arn=task_definition_arn, image_name=image_name, cluster_name=cluster_name)
        log_group_name  = config.get('log_group_name')
        log_stream_name = config.get('log_stream_name')
        return self.cloud_watch_logs().log_stream_exists(log_group_name=log_group_name, log_stream_name=log_stream_name)

    def logs(self, task_arn, task_definition_arn, image_name, cluster_name='default'):
        log_stream_config = self.log_stream_config(task_arn=task_arn, task_definition_arn=task_definition_arn, image_name=image_name, cluster_name=cluster_name)
        log_group_name    = log_stream_config.get('log_group_name')
        log_stream_name   = log_stream_config.get('log_stream_name')
        return self.cloud_watch_logs().log_events(log_group_name=log_group_name, log_stream_name=log_stream_name)

    def logs_wait_for_data(self, task_arn, task_definition_arn, image_name, cluster_name='default',max_attempts=60 ,sleep_for=1):
        for i in range(0, max_attempts):                                                                                                    # for max_attempts
            logs = self.logs(task_arn=task_arn,task_definition_arn=task_definition_arn,image_name=image_name,cluster_name=cluster_name)     # get logs (returns '' if cloud trail doesn't exist or has no data)
            if logs:                                                                                                                        # if there is data
                return logs                                                                                                                 # return it
            wait(sleep_for)                                                                                                                 # if not, wait for sleep_for amount

    def task_definition(self, task_definition_arn=None, task_family=None):
        try:
            if task_definition_arn is None:
                task_definition_arn = task_family
            if task_definition_arn:
                result          = self.client().describe_task_definition(taskDefinition=task_definition_arn)
                task_definition = result.get('taskDefinition')
                if task_definition and task_definition.get('status') == 'ACTIVE':       # threat inactive task definitions as non existent (to address AWS weird bug of not allowing task definitions to be deleted)
                    return task_definition
        except ClientError as e:
            if e.response.get('Error', {}).get('Message') != 'Unable to describe task definition.':
                raise
        return {}


    def task_definition_arn(self, task_family, revision):
        return f'arn:aws:ecs:{self.region}:{self.account_id}:task-definition/{task_family}:{revision}'

    def task_definition_create(self, task_definition_config, skip_if_exists=True):
        cpu                 = task_definition_config.get('cpu'                )
        execution_role_name = task_definition_config.get('execution_role_name')
        log_group_name      = task_definition_config.get('log_group_name'     )
        log_stream_name     = task_definition_config.get('log_stream_name'    )
        image_name          = task_definition_config.get('image_name'         )
        memory              = task_definition_config.get('memory'             )
        network_mode        = task_definition_config.get('network_mode'       )
        task_family         = task_definition_config.get('task_family'        )
        task_role_name      = task_definition_config.get('task_role_name'     )
        requires            = task_definition_config.get('requires'           )

        if skip_if_exists and self.task_definition_exists(task_family=task_family):
            return self.task_definition(task_family=task_family)

        # create policies if needed
        iam_task_role       = self.policy_create_for_task_role(task_role_name)
        iam_execution_role  = self.policy_create_for_execution_role(execution_role_name)
        task_role_arn       = iam_task_role.arn()
        execution_role_arn  = iam_execution_role.arn()

        self.cloud_watch_logs().log_stream_create(log_group_name=log_group_name, log_stream_name=log_stream_name)

        kwargs = { 'family'              : task_family           ,
                   'taskRoleArn'         : task_role_arn         ,
                   'executionRoleArn'    : execution_role_arn    ,
                   'cpu'                 : cpu                   ,
                   'memory'              : memory                ,
                   'networkMode'         : network_mode          ,
                   "containerDefinitions": [ { 'name'            : image_name,
                                               'image'           : f'{self.account_id}.dkr.ecr.{self.region}.amazonaws.com/{image_name}' ,
                                               "logConfiguration": {   "logDriver": "awslogs",
                                                                       "options"  : {  "awslogs-group"        : log_group_name,
                                                                                       "awslogs-region"       : self.region,
                                                                                       "awslogs-stream-prefix": log_stream_name }}, }],
                   "requiresCompatibilities": [ requires] }


        return self.client().register_task_definition(**kwargs).get('taskDefinition')

    def task_definition_delete(self, task_definition_arn):
        result = self.client().deregister_task_definition(taskDefinition=task_definition_arn)
        return result.get('taskDefinition').get('status') == 'INACTIVE'

    def task_definition_exists(self, task_definition_arn=None, task_family=None):
        return self.task_definition(task_definition_arn=task_definition_arn,task_family=task_family) != {}

    def task_definition_not_exists(self, task_definition_arn):
        return self. task_definition_exists(task_definition_arn=task_definition_arn) is False

    def task_definition_setup(self, task_family, image_name, cpu='256', memory='512', network_mode='awsvpc', requires='FARGATE'):
        task_role_name      = f'ecs-role-task__{self.region}__{task_family}'
        execution_role_name = f'ecs-role-execution__{self.region}__{task_family}'
        log_group_name      = f'ecs-task-on-{image_name}'

        return { 'cpu'                : cpu                  ,
                 'image_name'         : image_name           ,
                 'log_group_name'     : log_group_name       ,
                 'log_stream_name'    : task_family          ,
                 'memory'             : memory               ,
                 'execution_role_name': execution_role_name  ,
                 'network_mode'       : network_mode         ,
                 'requires'           : requires             ,
                 'task_family'        : task_family          ,
                 'task_role_name'     : task_role_name       }

    @index_by
    @group_by
    def task_definitions(self, task_family):
        data = []
        task_definitions_arns = self.task_definitions_arns(task_family)
        for task_definition_arn in task_definitions_arns:
            data.append(self.task_definition(task_definition_arn=task_definition_arn))
        return data

    def task_definitions_arns(self, task_family,status='ACTIVE'):
        kwargs = { 'familyPrefix': task_family ,
                    'status'     : status      }

        return self.client().list_task_definitions(**kwargs).get('taskDefinitionArns')

    def task(self, cluster_name, task_arn):
        result = self.client().describe_tasks(cluster=cluster_name, tasks=[task_arn]).get('tasks')
        return list_get(result,0)

    def task_create(self, task_definition_arn, subnet_id, security_group_id, cluster_name='default', launch_type='FARGATE', assign_public_ip='ENABLED', constraint_expression=None):

        kwargs = {
                    'cluster'             : cluster_name         ,
                    'taskDefinition'      : task_definition_arn  ,
                    'launchType'          : launch_type          ,
                    #'networkConfiguration': network_configuration,
                 }
        if launch_type != 'EXTERNAL':
            network_configuration          = {'awsvpcConfiguration': {'subnets': [subnet_id],
                                                                      'securityGroups': [security_group_id],
                                                                      'assignPublicIp': assign_public_ip}}
            kwargs['networkConfiguration'] = network_configuration
        if constraint_expression:
            kwargs['placementConstraints'] = [
                {
                    'type': 'memberOf',
                    'expression': constraint_expression
                }
            ]
        result = self.client().run_task(**kwargs)
        return result.get('tasks')[0]

    def task_create_ec2(self, task_definition_arn):
        kwargs = {"taskDefinition" : task_definition_arn ,
                  "launchType"     : "EC2"               ,
                  "count"          : 1                   }                                           # only create one instance

        result = self.client().run_task(**kwargs)
        return result
        #return result.get('tasks')[0]

    def task_exists(self, cluster_name, task_arn):
        return self.task(cluster_name=cluster_name, task_arn=task_arn) is not None

    def task_stop(self, cluster_name, task_arn):
        kwargs = {'cluster': cluster_name,
                  'task'   : task_arn    }

        return self.client().stop_task(**kwargs).get('taskDefinitionArns')

    # def task_wait_for_completion(self, task_arn, cluster_name='default', sleep_for=1, max_attempts=30, log_status=False):  # todo: replace with waiters that
    #     build_info = ''
    #     for i in range(0,max_attempts):
    #         build_info    = self.task(cluster_name=cluster_name, task_arn=task_arn)
    #         last_Status  = build_info.get('lastStatus')
    #         #current_phase = build_info.get('currentPhase')
    #         if log_status:
    #             Dev.pprint("[{0}] {1}".format(i,last_Status))
    #         if last_Status == 'DEPROVISIONING' or last_Status == 'STOPPED':
    #             return build_info
    #         sleep(sleep_for)
    #     return build_info

    @index_by
    @group_by
    def tasks(self, cluster_name, task_arns=None):
        if task_arns is None:
            task_arns = self.tasks_arns(cluster_name=cluster_name)
        kwargs = {"cluster": cluster_name,
                  "tasks"  : task_arns}
        return self.client().describe_tasks(**kwargs).get('tasks')

    def tasks_arns(self,cluster_name='default', task_family=None):
        all_tasks_arns= []
        all_tasks_arns.extend(self.tasks_arns_on_status(cluster_name=cluster_name, task_family=task_family, status= 'RUNNING'))
        all_tasks_arns.extend(self.tasks_arns_on_status(cluster_name=cluster_name, task_family=task_family, status= 'STOPPED'))
        return all_tasks_arns

    def tasks_arns_on_status(self, cluster_name, task_family=None, status='RUNNING'):
        kwargs = { "cluster"      : cluster_name,
                   'desiredStatus': status     }
        if task_family:
            kwargs['family'] = task_family
        return self.client().list_tasks(**kwargs).get('taskArns')

    def policy_create_for_task_role(self, role_name, skip_if_exists=True):
        policy = { "Version"    : "2008-10-17",
                   "Statement"  : [ { "Effect": "Allow",
                                      "Principal": { "Service": "ecs-tasks.amazonaws.com"},
                                      "Action": "sts:AssumeRole" }]}
        iam_role = IAM_Role(role_name=role_name)
        iam_role.create(policy,skip_if_exists=skip_if_exists)
        return iam_role


    def policy_create_for_execution_role(self, role_name, skip_if_exists=True):
        #cloud_watch_arn = "arn:aws:logs:{0}:{1}:log-group:awslogs-*".format(self.region, self.account_id)
        cloud_watch_arn = f"arn:aws:logs:{self.region}:{self.account_id}:log-group:*"
        role_document   = { "Version"   : "2008-10-17",
                            "Statement" : [ { "Effect": "Allow",
                                              "Principal": { "Service": "ecs-tasks.amazonaws.com"},
                                              "Action": "sts:AssumeRole" }]}
        policy_document = { "Version"  : "2012-10-17",
                            "Statement": [{   "Effect"  :  "Allow"                              ,
                                              "Action"  : [  "ecr:GetAuthorizationToken"        ,
                                                             "ecr:BatchCheckLayerAvailability"  ,
                                                             "ecr:GetDownloadUrlForLayer"       ,
                                                             "ecr:GetRepositoryPolicy"          ,
                                                             "ecr:DescribeRepositories"         ,
                                                             "ecr:ListImages"                   ,
                                                             "ecr:DescribeImages"               ,
                                                             "ecr:BatchGetImage"               ],
                                               "Resource": "*"                                 },
                                          {    "Effect"  : "Allow",
                                               "Action"  : [ "logs:CreateLogStream" ,
                                                             "logs:PutLogEvents"   ],
                                               "Resource": [ cloud_watch_arn ]}]}

        policy_name = 'policy_for_{0}'.format(role_name)

        iam_role = IAM_Role(role_name=role_name)

        if iam_role.exists() and skip_if_exists:
            return iam_role

        if iam_role.create(assume_policy_document=role_document, skip_if_exists=skip_if_exists):
            iam_role.attach_policy(policy_name=policy_name, policy_document=policy_document)
            if policy_name in iam_role.iam.role_policies():
                return iam_role

    def wait_for(self, waiter_type, kwargs):
        waiter = self.client().get_waiter(waiter_type)
        waiter.config.delay        = ECS_WAITER_DELAY
        waiter.config.max_attempts = ECS_WAITER_MAX_ATTEMPTS
        return waiter.wait(**kwargs)


    #todo: add other two waiters: wait_for_services_inactive and wait_for_services_stable

    def wait_for_task_running(self, task_arn, cluster_name='default'):
        waiter_type = 'tasks_running'
        kwargs      = { "cluster" : cluster_name,
                        "tasks"   : [task_arn  ]}
        self.wait_for(waiter_type=waiter_type, kwargs=kwargs)
        return self.task(cluster_name=cluster_name, task_arn=task_arn)

    def wait_for_task_stopped(self, task_arn, cluster_name='default'):
        waiter_type = 'tasks_stopped'
        kwargs      = { "cluster" : cluster_name,
                        "tasks"   : [task_arn  ]}
        self.wait_for(waiter_type=waiter_type, kwargs=kwargs)
        return self.task(cluster_name=cluster_name, task_arn=task_arn)


from osbot_utils.utils.Files import file_create, file_not_exists
from osbot_utils.utils.Process import start_process

def run(event, context=None):
    target_host    = event.get('target_host'   )
    ssh_key        = event.get('ssh_key'       )
    ssh_key_name   = event.get('ssh_key_name'  )
    ssh_user       = event.get('ssh_user'      )
    ssh_command    = event.get('ssh_command'   )
    # port_forward   = event.get('port_forwards' )              # not implemented:
    include_stderr = event.get('include_stderr')
    ssh_key_file   = f'/tmp/{ssh_key_name}'

    if file_not_exists(ssh_key_file):                           # create local key if it doesn't exist
        file_create(ssh_key_file, ssh_key)
        start_process('chmod', ['600', ssh_key_file])

    ssh_params = ['-o', 'StrictHostKeyChecking=no']             # todo: add support for updating the local hosts file
    if ssh_key_file:
        ssh_params.append('-i')                                 # set key to use
        ssh_params.append(ssh_key_file)
    # if port_forward:                                          # todo see if we do actually need this (main use case would be to allow direct HTTP access to an internal server)
    #     local_port  = port_forward.get('local_port' )         #      need to see if Lambda will allow binding ports like this
    #     remote_ip   = port_forward.get('remote_ip'  )
    #     remote_port = port_forward.get('remote_port')
    ssh_params.append(f'{ssh_user}@{target_host}')              # set user and target ip
    ssh_params.append(ssh_command)                              # add command to execute

    result = start_process("ssh", ssh_params)                   # execute command

    if include_stderr:                                          # see if we need to include stderr in return value
        return result.get('stdout') + result.get('stderr')
    return result.get('stdout')


from decimal import Decimal
from typing import Union

from boto3.dynamodb.types import TypeSerializer

from osbot_aws.decorators.enforce_type_safety import enforce_type_safety
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.decorators.methods.cache_on_self import cache_on_self
from osbot_utils.utils.Dev import pprint


class DyDB__Query__Builder(Kwargs_To_Self):
    table_name                  : str
    key_name                    : str
    key_value                   : str
    expression_attribute_names  : dict
    expression_attribute_values : dict
    update_expression           : str
    return_values               : str = 'NONE'
    kwargs                      : dict

    def build(self):
        self.kwargs['Key'                       ] = self.exp_key()
        self.kwargs['ReturnValues'              ] = self.return_values
        self.kwargs['TableName'                 ] = self.table_name
        self.kwargs['UpdateExpression'          ] = self.update_expression
        if self.expression_attribute_values:
            self.kwargs['ExpressionAttributeValues' ] = self.expression_attribute_values
        if self.expression_attribute_names:
            self.kwargs['ExpressionAttributeNames'  ] = self.expression_attribute_names

        return self.kwargs

    @enforce_type_safety
    def set__attribute_name(self, name: str, value: str):
        return self.set__attribute_names({name : value})

    @enforce_type_safety
    def set__attribute_names(self, names_dict: dict):
        self.expression_attribute_names.update(names_dict)
        return self

    @enforce_type_safety
    def set__attribute_value(self, key: str, value: any):
        return self.set__attribute_values({key : value})

    @enforce_type_safety
    def set__attribute_values(self, values_dict: dict):
        for key, value in values_dict.items():
            self.expression_attribute_values[key] = self.serialize_value(value)
        return self

    @enforce_type_safety
    def set__update_expression(self, command  : str, target       : str, operation: str = None,
                               condition: str = None, default_value: str = None, new_value: str = None):
        if command == 'SET':
            if operation and condition:                                                         # Construct the conditional part if a condition and operation are specified
                conditional_part = f"{condition}({target}, {default_value})"
                expression =  f"SET {target} = {operation}({conditional_part}, {new_value})"
            elif operation:                                                                     # For operations without a condition (direct setting or transformations)
                expression = f"SET {target} = {operation}({new_value})"
            else:
                expression = f"SET {target} = {new_value}"                                      # Simple SET operation
        elif command == 'REMOVE':
            expression = f'REMOVE {target}'
        elif command == 'DELETE':
            expression = f'DELETE {target}'
        elif command == 'ADD':
            expression = f'ADD {target}'
        else:
            expression = ''
        self.update_expression = expression

    @enforce_type_safety
    def build__add_to_list(self, list_field_name: str, new_list_element: any):
        target = '#list_field_name'
        self.set__attribute_name (name=target, value=list_field_name)
        self.set__attribute_value(key =':new_element', value=[new_list_element])
        self.set__attribute_value(key =":empty_list", value=[])
        expression_kwargs = dict(operation     = 'list_append'    ,
                                 condition     = 'if_not_exists'  ,
                                 default_value = ':empty_list',
                                 new_value     = ':new_element'   )
        self.set__update_expression('SET', target, **expression_kwargs)
        return self.build()

    def build__add_to_set(self, field_name: str, element: any):
        target = '#field_name'
        self.set__attribute_name (name=target, value=field_name)
        self.set__attribute_value(key =':element', value={element})
        self.set__update_expression(command='ADD', target='#field_name :element')
        return self.build()

    def build__dict_add_to_set(self, dict_field:str, field_name: str, element: any):
        target = '#field_name'
        self.set__attribute_names({'#dict_field': dict_field,
                                   '#field_name': field_name})
        self.set__attribute_values({':element'  : {element} })
        self.set__update_expression(command='ADD', target='#dict_field.#field_name :element')
        return self.build()


    @enforce_type_safety
    def build__dict_delete_field(self, dict_field: str, field_key: str) -> dict:
        self.set__attribute_names({ '#dict_field': dict_field   ,
                                    '#key'       : field_key    })
        self.set__update_expression(command='REMOVE', target='#dict_field.#key')
        return self.build()

    @enforce_type_safety
    def build__delete_from_set(self, dict_field: str, field_name: str, element: any):
        self.set__attribute_names  (names_dict  = { '#dict_field': dict_field,
                                                    '#field_name' : field_name })
        self.set__attribute_values (values_dict = {':elements': {element}       })
        self.set__update_expression(command     = "DELETE"                              ,
                                    target      = f'#dict_field.#field_name :elements' )
        return self.build()
    @enforce_type_safety
    def build__dict_set_field(self, dict_field: str, field_key: str, field_value: any) -> dict:
        self.set__attribute_names ({'#dict_field': dict_field  ,
                                    '#key'       : field_key})
        self.set__attribute_values({':val'       : field_value})
        self.set__update_expression(command   = 'SET',
                                    target    = '#dict_field.#key',
                                    new_value = ':val')
        return self.build()

    @enforce_type_safety
    def build__dict_dict_set_field(self, parent_dict_field: str, child_dict_field: str, field_key: str, field_value: any) -> dict:
        self.set__attribute_names({'#parent_dict_field': parent_dict_field,
                                   '#child_dict_field' : child_dict_field,
                                   '#key'              : field_key})
        self.set__attribute_values({':val'             : field_value})
        self.set__update_expression(command='SET',
                                    target='#parent_dict_field.#child_dict_field.#key',
                                    new_value=':val')
        return self.build()

    @enforce_type_safety
    def build__dict_update_counter(self, dict_name, field_name: str, increment_by: Union[int, Decimal]):
        self.set__attribute_names({'#field_name': field_name,
                                   '#dict_name' : dict_name})
        self.set__attribute_values({':inc': increment_by})
        self.set__update_expression(command='ADD', target='#dict_name.#field_name :inc')
        return self.build()

    @enforce_type_safety
    def build__delete_field(self, field_name: str) -> dict:
        self.set__attribute_names  (names_dict = {'#field_name': field_name})
        self.set__update_expression(command    = 'REMOVE'                   ,
                                    target     = '#field_name'              )
        return self.build()

    @enforce_type_safety
    def build__delete_item_from_list(self, list_field_name : str, item_index : Union[int, Decimal]):
        self.set__attribute_names  (names_dict = {'#list_field_name': list_field_name  })
        self.set__update_expression(command    = "REMOVE"                               ,
                                    target     = f'#list_field_name[{item_index}]'      )
        return self.build()

    @enforce_type_safety
    def build__delete_elements_from_set(self, set_field_name: str, elements: set):
        self.set__attribute_names  (names_dict  = { '#set_field_name' : set_field_name })
        self.set__attribute_values (values_dict = {':elementsToRemove': elements       })
        self.set__update_expression(command     = "DELETE"                              ,
                                    target      = f'#set_field_name :elementsToRemove' )
        return self.build()

    @enforce_type_safety
    def build__set_field(self, field_name: str, field_value: any) -> dict:
        self.set__attribute_names  ({'#field_name' : field_name  })
        self.set__attribute_values ({':field_value': field_value })
        self.set__update_expression(command   = "SET"            ,
                                    target    = '#field_name'    ,
                                    new_value = ':field_value'   )
        return self.build()

    @enforce_type_safety
    def build__increment_field(self, field_name:str, increment_by:Union[int, Decimal]):
        self.set__attribute_names  ({ '#field_name': field_name })
        self.set__attribute_values ({ ':inc': increment_by      })
        self.set__update_expression(command='ADD',target='#field_name :inc')
        return self.build()

    def exp_key(self):
        return {self.key_name: self.serialize_value(self.key_value) }

    def serialize_value(self, value):
        return self.type_serializer().serialize(value)

    @cache_on_self
    def type_serializer(self):
        return TypeSerializer()


from osbot_aws.aws.iam.IAM import IAM
from osbot_utils.utils.Misc import random_string


class Temp_IAM_User:
    def __init__(self):
        self.user_name  = f'temp_user_{random_string()}'
        self.iam        = IAM(user_name=self.user_name)

    def __enter__(self):
        self.iam.user_create()
        return self

    def __exit__(self, type, value, traceback):
        self.iam.user_delete()



import re
import botocore
from botocore.exceptions import ClientError, ParamValidationError
from functools import wraps
from osbot_utils.utils.Dev import pprint

class Capture_Boto3_Error:
    def __init__(self):
        self.error_details = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:                                                              # Check if an exception occurred
            if isinstance(exc_val, ClientError):                                # If it's a Boto3 ClientError, process it
                self.process_client_error(exc_val)
                return True
            elif isinstance(exc_val, ParamValidationError):                     # If it's a Boto3 ParamValidationError, process it
                self.process_param_validation_error(exc_val)
                return True
        return False  # Return False to allow any exceptions not handled here to propagate

    def process_client_error(self, exc_val):
        error_code = exc_val.response['Error']['Code']
        error_message = exc_val.response['Error']['Message']

        if error_code == 'ValidationException':
            self.error_details = {
                'error_code': error_code,
                'error_message': error_message,
                'details': self.extract_validation_details(error_message)
            }
        else:
            self.error_details = {
                'error_code': error_code,
                'error_message': error_message
            }

    def process_param_validation_error(self, exc_val):
        # Here you can parse exc_val's report attribute if needed
        error_message = str(exc_val)
        self.error_details = {
            'error_code': 'ParamValidationError',
            'error_message': error_message
        }

    @staticmethod
    def extract_validation_details(error_message):
        # Implement logic to extract details from the ValidationException message
        details_match = re.search(r"An error occurred \(ValidationException\) when calling the (.+) operation: (.+)", error_message)
        if details_match:
            operation, detail_message = details_match.groups()
            return {'operation': operation, 'detail_message': detail_message}
        return {'error_message': error_message}

def capture_boto3_error(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return_value = None
        with Capture_Boto3_Error() as error_capture:
            try:
                return_value = func(*args, **kwargs)
            except Exception as e:
                raise e
        if error_capture.error_details:
            print()
            pprint('****** BOTO3 ERROR DETECTED ******')
            pprint(error_capture.error_details)
        return return_value

    return wrapper



from osbot_utils.decorators.methods.cache import cache
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_aws.apis.EC2 import EC2


class VPC:
    def __init__(self, vpc_name, tags):
        self.internet_gateway_id  = None
        self.route_table_id       = None
        self.security_group_id    = None
        self.subnet_id            = None
        self.vpc_id               = None
        self.vpc_name             = vpc_name
        self.tags                 = tags

    @cache_on_self
    def ec2(self):
        return EC2()

    def create_internet_gateway(self):
        self.internet_gateway_id = self.ec2().internet_gateway_create(tags=self.tags).get('InternetGatewayId')
        self.ec2().vpc_attach_internet_gateway(vpc_id=self.vpc_id, internet_gateway_id=self.internet_gateway_id)

    def create_route_table(self):
        self.route_table_id = self.ec2().route_table_create(vpc_id=self.vpc_id, tags=self.tags).get('RouteTableId')

    def create_security_group(self):
        security_group_name    = self.vpc_name
        description            = self.vpc_name
        result                 = self.security_group_id = self.ec2().security_group_create(security_group_name=security_group_name, description=description, vpc_id=self.vpc_id,tags=self.tags)
        self.security_group_id = result.get('data').get('security_group_id')

    def create_subnet(self):
        self.subnet_id = self.ec2().subnet_create(vpc_id=self.vpc_id, tags=self.tags).get('SubnetId')

    def create_vpc(self):
        self.vpc_id = self.ec2().vpc_create(tags=self.tags).get('VpcId')

    def delete_internet_gateway(self):
        self.ec2().vpc_detach_internet_gateway(vpc_id=self.vpc_id, internet_gateway_id=self.internet_gateway_id)
        self.ec2().internet_gateway_delete(self.internet_gateway_id)

    def delete_security_group(self):
        self.ec2().security_group_delete(self.security_group_id)

    def delete_subnet(self):
        self.ec2().subnet_delete(self.subnet_id)

    def delete_route_table(self):
        self.ec2().route_table_delete(self.route_table_id)

    def delete_vpc(self):
        self.ec2().vpc_delete(self.vpc_id)


import os
from osbot_aws.aws.s3.S3                            import S3
from osbot_aws.AWS_Config                           import AWS_Config
from osbot_utils.decorators.methods.cache_on_self   import cache_on_self
from osbot_utils.utils.Files                        import Files, folder_exists, folder_not_exists, folder_create, file_not_exists
from osbot_utils.utils.Process                      import Process

# todo: move to lambdas_ section (since this is used by lambdas to dynamically load dependencies from zip files in s3)

class Dependencies:

    @cache_on_self
    def s3(self):
        return S3()

    def load_dependencies(self, targets):
        if type(targets) is list:
            for target in targets:
                self.load_dependency(target)
            return
        for target in targets.split(','):
            self.load_dependency(target.strip())

    def load_dependency(self, target):
        if os.getenv('AWS_REGION') is None:
            return

        import shutil
        import sys
        s3         = self.s3()
        s3_bucket  = AWS_Config().lambda_s3_bucket()
        s3_key     = 'lambdas-dependencies/{0}.zip'.format(target)
        tmp_dir    = Files.path_combine('/tmp/lambdas-dependencies', target)
        #return s3.file_exists(s3_bucket,s3_key)

        if s3.file_exists(s3_bucket,s3_key) is False:
            raise Exception("In Lambda load_dependency, could not find dependency for: {0}".format(target))

        if file_not_exists(tmp_dir):                                # download dependency
            zip_file = s3.file_download(s3_bucket, s3_key,False)    # download zip file with dependencies
            shutil.unpack_archive(zip_file, extract_dir = tmp_dir)  # unpack them
        if tmp_dir not in sys.path:                                 # if not currently in the path
            sys.path.append(tmp_dir)                                # add tmp_dir to the path that python uses to check for dependencies
        return Files.exists(tmp_dir)

    def pip_install_dependency(self, target, target_aws_lambda=True):
        path_lambda_dependencies = Files.path_combine('.', '../../../_lambda_dependencies/')
        folder_create(path_lambda_dependencies)
        path_install = Files.path_combine(path_lambda_dependencies, target)
        if folder_not_exists(path_install):
            args = ['install']
            if target_aws_lambda:
                args.extend(['--platform','manylinux1_x86_64', '--only-binary=:all:'])
            args.extend(['-t', path_install, target])
            return Process.run('pip3', args)
        return folder_exists(path_install)

    def upload_dependency(self, target):
        s3        = self.s3()
        s3_bucket = AWS_Config().lambda_s3_bucket()
        s3_file   = 'lambdas-dependencies/{0}.zip'.format(target)
        path_libs = Files.path_combine('../../../_lambda_dependencies/', target)
        if Files.not_exists(path_libs):
            raise Exception(f"In Lambda upload_dependency, could not find dependency for: {target} , which resolved to {path_libs}")
        s3.folder_upload(path_libs, s3_bucket, s3_file)
        return s3.file_exists(s3_bucket, s3_file)

# todo refactor into class (see Lambda_Upload_Package})

# LEGACY METHODS
def load_dependencies(targets):
    Dependencies().load_dependencies(targets)

def load_dependency(target):
    Dependencies().load_dependency(target)

def pip_install_dependency(target, target_aws_lambda=True):
    Dependencies().pip_install_dependency(target, target_aws_lambda)

def upload_dependency(target):
    Dependencies().upload_dependency(target)



from unittest import TestCase

from osbot_aws.aws.apigateway.API_Gateway_V2 import API_Gateway_V2
from osbot_aws.aws.apigateway.API_Gateway_V2__with_temp_role import API_Gateway_V2__with_temp_role
from osbot_aws.testing.Pytest import skip_pytest___aws_pytest_user_name__is_not_set


class TestCase__API_Gateway_V2(TestCase):
    api_gateway_v2  : API_Gateway_V2
    reset_iam_creds : bool = False

    @classmethod
    def setUpClass(cls):
        skip_pytest___aws_pytest_user_name__is_not_set()
        cls.api_gateway_v2 = API_Gateway_V2__with_temp_role()
        if cls.reset_iam_creds:
            cls.api_gateway_v2.temp_role__iam_reset_credentials()


    @classmethod
    def tearDownClass(cls):
        pass


from osbot_aws.aws.iam.IAM_Role import IAM_Role


class Temp_IAM_Role:
    def __init__(self, role_name=None, delete_on_exit=True):
        self.role_name      = role_name
        self.iam_role       = IAM_Role(role_name=role_name)
        self.delete_on_exit = delete_on_exit


    def __enter__(self):
        self.iam_role.create_for__lambda()
        return self.iam_role

    def __exit__(self, type, value, traceback):
        if self.delete_on_exit:
            self.iam_role.delete()



from osbot_utils.utils.Process import Process


class Lambda_Docker:

    def __init__(self):
        self.docker_image = "lambci/lambda:python3.8"


    def run_bash(self, cmd):
        run_params = ["run", "--rm", "--entrypoint", "bash", self.docker_image, "-c",cmd]
        return Process.run("docker", run_params)

    # todo: finish this method
    # def invoke(self, cmd):
    #     run_params = ["run", "--rm",
    #                   "--entrypoint", "bash",
    #                   self.docker_image, "-c", cmd]
    #     # docker run --rm                             \

        #           -v "$PWD":/var/task:ro,delegated \
        #           lambci/lambda:python3.8          \
        #           lambda_function.run




from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion import Stability_Stable_Diffusion
from osbot_aws.aws.bedrock.models.diffusion.Stability_Stable_Diffusion_XL_V1 import Stability_Stable_Diffusion_XL_V1
from osbot_utils.base_classes.Kwargs_To_Self import Kwargs_To_Self
from osbot_utils.utils.Png import save_png_base64_to_file


class Stability_Stable_Diffusion_XL_V1__Image_To_Image(Stability_Stable_Diffusion_XL_V1):
    init_image      : str                           # The base64 encoded image that you want to use to initialize the diffusion process.
    init_image_mode : str     = 'IMAGE_STRENGTH'    # (Optional) Determines whether to use image_strength or step_schedule_* to control how much influence the image in init_image has on the result. Possible values are IMAGE_STRENGTH or STEP_SCHEDULE.
    image_strength  : float   = 1.0                 # (Optional) Determines how much influence the source image in init_image has on the diffusion process. Values close to 1 yield images very similar to the source image. Values close to 0 yield images very different than the source image.

    def body(self):
        return dict(**super().body()                                   ,
                    init_image      = self.init_image      ,
                    init_image_mode = self.init_image_mode ,
                    strength        = self.image_strength  )





# todo: refactor code to use DyDB__Table_With_Timestamp which has improved support for using timestamps in DyDB tables
# from os import environ
#
# from osbot_aws.aws.dynamo_db.Dynamo_DB__Table   import Dynamo_DB__Table
# from osbot_aws.aws.dynamo_db.domains.DyDB__Table import DyDB__Table
# from osbot_utils.utils.Misc                     import timestamp_utc_now
# from osbot_utils.utils.Objects import type_full_name
#
# NAME_TIMESTAMP   = 'timestamp'
# NAME_ENVIRONMENT = 'environment'
# NAME_DATA        = 'data'
# VALUE_NA         = 'NA'
#
# class DyDB__Timeseries(DyDB__Table):
#
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.global_secondary_indexes = []
#
#     def add_document(self, document, env=None):
#         if NAME_ENVIRONMENT not in document:
#             document[NAME_ENVIRONMENT] = environ.get('EXECUTION_ENV') or VALUE_NA
#         if NAME_TIMESTAMP not in document:
#             document[NAME_TIMESTAMP] = timestamp_utc_now()
#         result = super().add_document(document)
#         return result
#
#     def all_ids(self):                                                                          # todo: see if this can be moved to Dynamo_DB__Table
#         response = self.dynamo_db.client().scan(TableName            = self.table_name ,
#                                                 ProjectionExpression = self.key_name   )
#         ids = []
#         for item in response.get('Items'):
#             ids.append(item.get(self.key_name).get('S'))
#         return ids
#
#     def clear_table(self):                                                                      # todo: see if this can be moved to Dynamo_DB__Table
#         table_name  = self.table_name
#         key_name    = self.key_name
#         keys_values = self.all_ids()
#         delete_result = self.dynamo_db.documents_delete(table_name=table_name, key_name=key_name, keys_values=keys_values)
#         return delete_result
#
#     def query_by_partition(self, partition_value):
#         table_name              = self.table_name
#         index_name              = self.index_name
#         partition_key_name      = self.partition_key_name
#         primary_key_placeholder = "#pk"
#
#         kwargs = dict(TableName                 = table_name,
#                       IndexName                 = index_name,
#                       KeyConditionExpression    = f'{primary_key_placeholder} = :value' ,
#                       ExpressionAttributeNames  = { primary_key_placeholder: partition_key_name},
#                       ExpressionAttributeValues = { ':value': {'S': partition_value  }})
#         result = self.query(**kwargs)
#         items  = []
#         if result:
#             for raw_item in result.get('data').get('Items'):
#                 item = self.dynamo_db.document_deserialise(raw_item)
#                 items.append(item)
#         return items
#
#     def query_by_last_n_hours(self, hours, partition=None):
#         start_timestamp = timestamp_utc_now() - (hours * 60 * 60 * 1000)
#         end_timestamp   = timestamp_utc_now()
#         return self.query_by_timestamp(start_timestamp, end_timestamp, partition=partition)
#
#     def query_by_timestamp(self, start_timestamp, end_timestamp, partition=None):
#         if partition == 'None':
#             partition = self.partition_key_value
#         table_name              = self.table_name
#         index_name              = self.index_name
#         key_name                = self.key_name
#         partition_key_name      = self.partition_key_name
#         partition_key_value     = partition or self.partition_key_value
#         primary_key_placeholder = "#pk"
#         sort_key_placeholder    = "#sk"
#
#         kwargs = dict(TableName                 = table_name,
#                       IndexName                 = index_name,
#                       KeyConditionExpression    = f'{primary_key_placeholder} = :value AND {sort_key_placeholder} BETWEEN :start AND :end',
#                       ExpressionAttributeNames  = { sort_key_placeholder   : key_name          ,
#                                                     primary_key_placeholder: partition_key_name},
#                       ExpressionAttributeValues = { ':value': {'S': partition_key_value  },
#                                                     ':start': {'N': str(start_timestamp) },
#                                                     ':end'  : {'N': str(end_timestamp  ) }})
#         result = self.query(**kwargs)
#         items  = []
#         for raw_item in result.get('data', {}).get('Items',[]):
#             item = self.dynamo_db.document_deserialise(raw_item)
#             items.append(item)
#         return items
#
#     def query_by_primary_key(self, key_value):
#         table_name                  = self.table_name
#         dynamodb_client             = self.dynamo_db.client()
#         key_condition_expression    = '#primary_key_name = :primary_key_value'
#         expression_attribute_names  = { '#primary_key_name': self.primary_key }
#         expression_attribute_values = { ':primary_key_value': {'S': key_value} }
#         result = dynamodb_client.query( TableName                 = table_name                  ,
#                                         KeyConditionExpression    = key_condition_expression    ,
#                                         ExpressionAttributeNames  = expression_attribute_names  ,
#                                         ExpressionAttributeValues = expression_attribute_values )
#
#         items = result.get('Items', [])
#         return self.dynamo_db.document_deserialise(items[0]) if items else {}